#include "Arbre.h"

//SDL_Color structures
SDL_Color black = {0x00,0x00,0x00,0};
SDL_Color red = {0xFF,0x00,0x00,0};
SDL_Color blue = {0x00,0x00,0xFF,0};
SDL_Color green = {0x00,0xFF,0x00,0};
SDL_Color light_blue = {0x00,0xFF,0xFF,0};
SDL_Color purple = {0xFF,0x00,0xFF,0};
SDL_Color yellow = {0xFF,0xFF,0x00,0};
SDL_Color grey = {0x99,0x99,0x99,0};
SDL_Color white = {0xFF,0xFF,0xFF,0};

// visualisation 2D de l'arbre construit; dans un fichier, puis en SDL
void Arbre::visualize() {
	/***
	fichier
	***/

	ofstream vizFile("results/arbre.txt");
	vector<Noeud*> aVisiter;
	aVisiter.push_back(racine);
	Noeud* node;
	while (!aVisiter.empty()) {
		node=aVisiter.back();
		// affichage
		vizFile << "noeud type " << node->type;
		if (node->type!='f') vizFile << " et index att: " << node->indexAtt << endl;
		else vizFile << endl;
		if (node->type=='d') {
			for (int i=0; i<(int)node->subsets.size(); i++) {
				vizFile << "ensemble " << i << ":";
				for (int j=0; j<(int)node->subsets[i].size(); j++) vizFile << " " << node->subsets[i][j];
				vizFile << endl;
			}
		}
		else if (node->type=='c')
			vizFile << "point de coupure: " << node->cutFloat << endl;
		else if (node->type=='o')
			vizFile << "point de coupure: " << node->cutInt << endl;
		else 
			vizFile << "feuille, label == " << node->label << endl;
		vizFile << endl;
		// mise a jour de aVisiter
		aVisiter.pop_back();
		for (int i=node->nbFils-1; i>=0; i--) aVisiter.push_back(node->fils[i]);
	}

	/***
	en SDL
	***/

	//used surfaces
	SDL_Surface *screen;
	SDL_Surface *screen_backup;

	//the system video is initialized 
	if( SDL_Init(SDL_INIT_VIDEO) <0 ) {
		//SDL_GetError() returns a description of the error
		printf("SDL Init Error: %s\n", SDL_GetError());
	        return;
	}

	//when exit, execute SDL_Quit to restore everything
	atexit(SDL_Quit);

	//a 32 bit integer used to store screen's flags 
	Uint32 flags = SDL_HWSURFACE|SDL_DOUBLEBUF;

	//SDL_SetVideoMode is used to have screen associated to the monitor 
	if(!( screen = SDL_SetVideoMode(DIM_H, DIM_V, 0,flags) )) {
		  printf("Screen Init Error : %s\n", SDL_GetError());
		  return;
	}

	//white background
	fill_square(screen, NULL, white);

	//screen_backup is a surface of the same dimensions of screen used  to save the image when the user
	//changes its form (window to full screen and vice versa). 
	screen_backup = SDL_CreateRGBSurface(flags,screen->w,screen->h,screen->format->BitsPerPixel,0,0,0,0);

	//d'abord on calcule la profondeur de l'arbre pour definir l'ecart en pixels entre deux etages
	int prof=1; // 1 pour la racine
	vector<Noeud*> nodesToExplore;
	Noeud* curNode; // noeud en cours d'exploration
	racine->prof=1;
	nodesToExplore.push_back(racine);
	while (!nodesToExplore.empty()) {
		curNode=nodesToExplore.back();
		nodesToExplore.pop_back();
		if (curNode->type=='f') { 
			// cas d'arret: feuille
			if (curNode->prof>prof) prof=curNode->prof;
		}
		else {
			// on ajoute les fils qui prennent comme profondeur curNode->prof+1
			for (int i=curNode->nbFils-1; i>=0; i--) {
				curNode->fils[i]->prof=curNode->prof+1;
				nodesToExplore.push_back(curNode->fils[i]);
			}
		}
	}

	//puis on calcule les largeurs de chaque etage pour definir l'ecart horizontal entre les noeuds
	//au passage, on etiquete facilement les abscisses des noeuds pour faciliter le calcul futur
	//c'est possible grace au parcours en profondeur "de gauche a droite" realise
	largeurs=(int*)calloc(prof,sizeof(int)); // les largeurs des differentes profondeurs
	nodesToExplore.push_back(racine);
	while (!nodesToExplore.empty()) {
		curNode=nodesToExplore.back();
		nodesToExplore.pop_back();
		curNode->x=largeurs[curNode->prof-1]++;
		// on ajoute les fils dans l'ensemble a explorer
		for (int i=curNode->nbFils-1; i>=0; i--) nodesToExplore.push_back(curNode->fils[i]);
	}

	//on en deduit enfin la taille d'une cellule et les ecarts entre cellules
	//systeme d'inequations a resoudre "raisonnablement":
	// "prof*cell_side+(prof-1)*ecartVertical<DIM_V"
	// transforme en: prof*(cell_side+ecartVertical)<=DIM_V (pour laisser une petite marge en haut et en bas)
	// "larg*cell_side+(larg-1)*ecartHorizontal<DIM_H"
	//	transforme en: larg*(cell_side+ecartHorizontal)<=DIM_H (pour laisser une petite marge a droite et a gauche)

	// idealement, ecartVertical=1.5*cell_side (visuellement parlant)
	// si ce n'est pas possible, on determine alors les rapports pour essayer d'avoir cell_side>=10
	int dimCellV;
	if (DIM_V/(2.5*prof)>=10) {
		dimCellV=(int)(DIM_V/(2.5*prof));
		ecartV=(int)(1.5*dimCellV);
	}
	else {
		if (DIM_V/prof-10>=10) {
			dimCellV=10;
			ecartV=DIM_V/prof-10;
		}
		else if (DIM_V/prof-5>=5) { // test a 5, avant le pixel
			dimCellV=5;
			ecartV=DIM_V/prof-5;
		}
		else {
			dimCellV=1;
			ecartV=DIM_V/prof-1;
		}
	}

	// la dimension d'une cellule et le minimum des valeurs calculees verticalement, horizontalement et de 50 (arbitraire)
	int larg=0;
	for (int i=prof-1; i>=0; i--)
		if (largeurs[i]>larg) larg=largeurs[i];
	cell_side=min(50,dimCellV,DIM_H/(larg+1));
	ecartV=min(75,ecartV); // on ne laisse pas plus de 75 pixels d'intervalle, au cas ou

	// si l'ecart vertical calcule est trop petit, ou la largeur trop importante, ou la cellule trop petite,
	// on ne trace pas l'arbre
	if (cell_side<=0 || ecartV<3 || (DIM_H/larg-cell_side<0)) {
		vizFile << endl << "***********************" << endl << "arbre trop complexe pour trace" << endl << "***********************" << endl;
		return;
	}
	
	//on peut alors dessiner l'arbre
	draw_tree(screen);
	
	//SDL_Event struct used to manage events
	SDL_Event event;

	//variables used to control the game loop 
	bool done = false;
	bool paused = false;

	//game or main loop (as you prefer)
	while(!done) {
		//SDL_PollEvent manages user's events 
		while(SDL_PollEvent(&event)) {
			
			//the program quits if someone press the x symbol of the window with the mouse 
			if(event.type == SDL_QUIT)  
				done = true;  

			if(event.type == SDL_KEYDOWN) {
				//it quits also if someone press ESC 
				if ( event.key.keysym.sym == SDLK_ESCAPE ) 
					done = true; 
				else if ( event.key.keysym.sym == SDLK_p ) 
					paused = !paused ;

				//visualization form is changed if someone press SPACE 
				//(from window to full screen and vice versa)
				if(event.key.keysym.sym == SDLK_SPACE ) {
					//the actual screen is saved in screen_backup
					SDL_BlitSurface(screen,NULL,screen_backup,NULL);
					//visualization form is changed
					flags^=SDL_FULLSCREEN;
					screen = SDL_SetVideoMode(DIM_H, DIM_V, 0,flags);
					//the original image is reBlitted
					SDL_BlitSurface(screen_backup,NULL,screen,NULL);
				}
			}
		}

		//the screen is refreshed
		SDL_Flip(screen);

		//a minimum delay 
		SDL_Delay(1);
	}
}

void Arbre::draw_tree(SDL_Surface *surface) {
	// suivant le nombre de classes, on les indique par une couleur ou par un numero
	// le blanc est utilise dans le cas des couleurs pour representer une feuille sans exemples
	// pour l'instant ca n'arrive pas, mais pourrait survenir dans une extension future du programme
	SDL_Color* couleurs=new SDL_Color[7];
	bool color=false;
	if ((int)(racine->distribExemples.size())<8) {
		// on assigne une couleur a chaque id de classe
		couleurs[0]=red;
		couleurs[1]=blue;
		couleurs[2]=green;
		couleurs[3]=light_blue;
		couleurs[4]=purple;
		couleurs[5]=yellow;
		couleurs[6]=grey;
		color=true;
	}

	// on parcourt l'arbre pour savoir ou dessiner les noeuds + liens de parente
	vector<Noeud*> nodesToDraw;
	Noeud* curNode; // noeud en cours de tracage
	SDL_Rect cell = {0,0,cell_side,cell_side};
	point p,f;
	racine->y=1;
	nodesToDraw.push_back(racine);
	while (!nodesToDraw.empty()) {
		curNode=nodesToDraw.back();
		nodesToDraw.pop_back();
		// on dessine le noeud courant a sa position calculee:
		if (curNode==racine) cell.x=DIM_H/2;
		else cell.x=(DIM_H-(largeurs[curNode->prof-1]-1)*(DIM_H/(largeurs[curNode->prof-1]+1)))/2+(curNode->x)*(DIM_H/(largeurs[curNode->prof-1]+1));
		cell.y=curNode->y;
		
		if (curNode->type=='f') {
			if (color && !curNode->distribExemples.empty()) {
				for (int i=curNode->distribExemples.size()-1; i>=0; i--) {
					// test necessaire dans le cas de feuilles "impures"
					if (curNode->distribExemples[i].str==curNode->label) {
						draw_square(surface, &cell, black, couleurs[curNode->distribExemples[i].index]);
						break;
					}
				}
			}				
			else draw_square(surface, &cell, black, white);
		}
		else {
			p.x=cell.x+cell_side/2; // abscisse du centre du pere
			p.y=cell.y+cell_side/2; // ordonnee du meme centre
			draw_square(surface, &cell, black, black);
			for (int i=curNode->nbFils-1; i>=0; i--) {
				curNode->fils[i]->y=curNode->y+cell_side+ecartV;
				// enfin on trace le lien pere-fils:
				f.x=(DIM_H-(largeurs[curNode->fils[i]->prof-1]-1)*(DIM_H/(largeurs[curNode->fils[i]->prof-1]+1)))/2
					+(curNode->fils[i]->x)*(DIM_H/(largeurs[curNode->fils[i]->prof-1]+1))+cell_side/2; // abscisse du centre du fils i
				f.y=curNode->fils[i]->y+cell_side/2; // ordonnee du meme centre
				draw_line(surface, p, f, black);
				nodesToDraw.push_back(curNode->fils[i]);
			}
		}
	}
	delete [] couleurs;
	free(largeurs);
}
