#ifndef _Arbre
#define _Arbre

#include <vector>
#include <fstream>
#include <iostream>
#include "Graphic.h"
#include "SDL/SDL.h"
#include "Noeud.h"

using namespace std;

// classe pour representer et visualiser l'arbre calcule via la librairie SDL
// les noeuds internes de l'arbre sont etiquetes par:
// - un ensemble discret dans le cas d'un attribut symbolique
// - une valeur reelle dans le cas d'un attribut continu (separation de l'espace en 2,
//		on ne considere pas le cas de formules plus complexes)
// - une valeur discrete dans le cas d'un ensemble discret ordonne

class Arbre {

public:
	
	Arbre () {
		largeurs=0;
	}
	
	// constructeur placant la racine, son indice de discrimination encore inconnu
	Arbre (vector<StringIndexCounter>& nbExemplesClasses) {
		racine=new Noeud(nbExemplesClasses);
		openNodes.push_back(racine);
		largeurs=0;
	}
	
	~Arbre () {
		if (largeurs!=0) free(largeurs);
	}
	
	// ajout d'une feuille non resolue
	void addLeaf(Noeud* node) {
		openNodes.push_back(node);
	}
	
	// on choisit un noeud non encore resolu que l'on va traiter (il sera alors un noeud interne)
	bool getOpenNode(Noeud** node) {
		if (openNodes.empty()) return false;
		*node=openNodes.back();
		openNodes.pop_back();
		return true;
	}
	
	// fonctions min auxiliaires
	int min(int a,int b) {
		if (a<b) return a;
		else return b;
	}
	int min(int a,int b,int c) {
		if (a<b) {
			if (a<c) return a;
			else return c;
		}
		else {
			if (b<c) return b;
			else return c;
		}
	}
	
	// visualisation fichier + 2D de l'arbre construit
	void visualize();
	void draw_tree(SDL_Surface *surface); // dessin de l'arbre sur une surface

	Noeud* racine; // le noeud racine a partir duquel tout l'arbre est construit
	vector<Noeud*> openNodes; // noeuds non encore resolus (nbPos>0 et nbNeg>0)
	
	// variables utiles au graphisme
	int* largeurs;
	int cell_side,ecartV;
	
	// constantes utiles dans Prune, Test et Classif
	static constexpr float PRUNING_FACTOR=1.0; // plus il est eleve plus on coupe (cas !fileTest)
	static constexpr float WEIGHT_ERROR=0.5; // plus il est eleve plus on evite les erreurs, 
										// et donc moins on coupe (cas fileTest)
	static constexpr float PRUNING_FILTER=0.6; // seuil au-dela duquel on n'essaye meme pas de fusionner
	static constexpr int NBMIN_REPDISC=7; // nombre minimal d'exemples necessaires pour continuer 
										// la classification quand attribut symbolique manquant
	static constexpr int NBMIN_REPBIN=10; // nombre minimal d'exemples necessaires pour continuer 
										// la classification quand attribut cont/int manquant

};

#endif
