#include "ClassifBayes.h"

// trouve les parametres des gaussiennes (supposees independantes dans ce modele)
void ClassifBayes::findParams() {
	// calcul des probabilites des classes:
	classesProba=(double*)malloc(nbClasses*sizeof(double));
	int* repartitions=(int*)calloc(nbClasses,sizeof(int));
	for (int i=nbLines_A-1; i>=0; i--) repartitions[classIndex_A[i]]++;
	for (int i=nbClasses-1; i>=0; i--) classesProba[i]=(double)(repartitions[i]+1)/nbLines_A;
	free(repartitions);

	// calcul des moyennes puis ecarts-types
	means=(double**)malloc(nbClasses*sizeof(double*));
	stdDev=(double**)malloc(nbClasses*sizeof(double*));
	repartitions=(int*)malloc(nbSymbols*sizeof(int));
	distribSymb=(double***)malloc(nbClasses*sizeof(double**));
	int count,sumRep;
	for (int i=nbClasses-1; i>=0; i--) {
		means[i]=(double*)calloc(nbAtt,sizeof(double));
		stdDev[i]=(double*)calloc(nbAtt,sizeof(double));
		distribSymb[i]=(double**)malloc(attSymbolic_A.size()*sizeof(double*));
		for (int j=typesAttributs.size()-1; j>=0; j--) {
			count=0;
			switch (typesAttributs[j]) {
				case 's':
					distribSymb[i][tabIndex[j]]=(double*)calloc(nbSymbols,sizeof(double));
					for (int k=nbSymbols-1; k>=0; k--) repartitions[k]=0;
					sumRep=0;
					for (int k=nbLines_A-1; k>=0; k--) {
						if (!attSymbolic_A[tabIndex[j]][k].miss && classes_A[k]==initData[i].str) {
							repartitions[indexOfSymbol_A[tabIndex[j]][k]]++;
							sumRep++;
						}
					}
					if (sumRep>0) {
						for (int k=nbSymbols-1; k>=0; k--)
							distribSymb[i][tabIndex[j]][k]=(double)repartitions[k]/sumRep;
					}
					break;
				case 'c':
					for (int k=nbLines_A-1; k>=0; k--) {
						if (!attContinue_A[tabIndex[j]][k].miss && classes_A[k]==initData[i].str) {
							means[i][j]+=new_attContinue_A[tabIndex[j]][k];
							stdDev[i][j]+=new_attContinue_A[tabIndex[j]][k]*new_attContinue_A[tabIndex[j]][k];
							count++;
						}
					}
					break;
				case 'd':
					for (int k=nbLines_A-1; k>=0; k--) {
						if (!attDiscrete_A[tabIndex[j]][k].miss && classes_A[k]==initData[i].str) {
							means[i][j]+=new_attDiscrete_A[tabIndex[j]][k];
							stdDev[i][j]+=new_attDiscrete_A[tabIndex[j]][k]*new_attDiscrete_A[tabIndex[j]][k];
							count++;
						}
					}
					break;
				default: break; // ce cas inclue les identifiants 'i'
			}
			if (count>0) {
				means[i][j]/=count;
				stdDev[i][j]=sqrt(stdDev[i][j]/count-means[i][j]*means[i][j]);
			}
		}
	}
	free(repartitions);
}

// classifie l'exemple de la ligne numLine, supposant l'appel a la fonction findParams() realise
string& ClassifBayes::classify(int numLine,bool testData) {
	double prodProba,bestProba=-1.0;
	int bestIndex=-1;
	for (int i=nbClasses-1; i>=0; i--) {
		// on effectue le produit des probabilites (en supposant l'independance)
		prodProba=classesProba[i];
		for (int j=nbAtt-1; j>=0; j--) {
			if (typesAttributs[j]!='s' && stdDev[i][j]==0.0) continue;
			switch (typesAttributs[j]) {
				case 's':
					if (testData && !attSymbolic_T[tabIndex[j]][numLine].miss
						&& distribSymb[i][tabIndex[j]][indexOfSymbol_T[tabIndex[j]][numLine]]>0.0)
						prodProba*=distribSymb[i][tabIndex[j]][indexOfSymbol_T[tabIndex[j]][numLine]];
					else if (!testData && !attSymbolic_A[tabIndex[j]][numLine].miss
						&& distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][numLine]]>0.0)
						prodProba*=distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][numLine]];
					break;
				case 'c':
					if (testData && !attContinue_T[tabIndex[j]][numLine].miss)
						prodProba*=(1.0/stdDev[i][j])*exp(-0.5*square((new_attContinue_T[tabIndex[j]][numLine]-means[i][j])/stdDev[i][j]));
					else if (!testData && !attContinue_A[tabIndex[j]][numLine].miss)
						prodProba*=(1.0/stdDev[i][j])*exp(-0.5*square((new_attContinue_A[tabIndex[j]][numLine]-means[i][j])/stdDev[i][j]));
					break;
				case 'd':
					if (testData && !attDiscrete_T[tabIndex[j]][numLine].miss)
						prodProba*=(1.0/stdDev[i][j])*exp(-0.5*square((new_attDiscrete_T[tabIndex[j]][numLine]-means[i][j])/stdDev[i][j]));
					else if (!testData && !attDiscrete_A[tabIndex[j]][numLine].miss)
						prodProba*=(1.0/stdDev[i][j])*exp(-0.5*square((new_attDiscrete_A[tabIndex[j]][numLine]-means[i][j])/stdDev[i][j]));
					break;
				default: break;
			}
		}
		if (prodProba>bestProba) {
			bestProba=prodProba;
			bestIndex=i;
		}
	}

	return initData[bestIndex].str;
}

// affichage des parametres dans un fichier
// remarque: les resultats de calculs avec valeurs inutiles (identifiants) sont simplement a zero
void ClassifBayes::printToFile() {
	ofstream outFile("results/paramBayes.txt");
	bool meansPrinted;
	outFile << "/// probabilites des classes: ///" << endl << endl;
	for (int i=0; i<nbClasses; i++) outFile << initData[i].str << "  " << classesProba[i] << endl;
	outFile << endl << "/// moyennes et ecarts-types des attributs numeriques, et distributions des attributs symboliques: ///" << endl << endl;
	for (int i=0; i<nbClasses; i++) {
		meansPrinted=false;
		outFile << "*** " << initData[i].str << " ***" << endl << endl;
		for (int j=0; j<nbAtt; j++) {
			if (typesAttributs[j]=='c' || typesAttributs[j]=='d') {
				if (!meansPrinted) {
					outFile << "MEAN=" << endl << "   ";
					meansPrinted=true;
				}
				outFile << fixed << means[i][j] << " ";
			}
		}
		if (meansPrinted) {
			outFile << endl;
			outFile << "STD_DEV=" << endl;
			for (int j=0; j<nbAtt; j++) {
				if (typesAttributs[j]=='c' || typesAttributs[j]=='d')
					outFile << fixed << stdDev[i][j] << " ";
			}
		}
		if (!attSymbolic_A.empty()) {
			if (meansPrinted) outFile << endl;
			outFile << "les probabilites des symboles respectifs suivants: " << endl;
			for (int j=nbSymbols-1; j>=0; j--) outFile << symbols[j] << "  ";
			outFile << endl << "sont, par attribut:" << endl;
		}
		for (int j=0; j<(int)attSymbolic_A.size(); j++) {
			outFile << (j+1) << ":" << endl;
			for (int k=nbSymbols-1; k>=0; k--) outFile << distribSymb[i][j][k] << " ";
			outFile << endl;
		}
		outFile << endl;
		if (!attSymbolic_A.empty()) outFile << endl;
	}
}
