#include "ClassifGauss.h"

// trouve les parametres des gaussiennes
void ClassifGauss::initialize() {
	// calcul des probabilites des classes:
	probasClasses=(double*)malloc(nbClasses*sizeof(double));
	int* repartitions=(int*)calloc(nbClasses,sizeof(int));
	for (int i=nbLines_A-1; i>=0; i--) repartitions[classIndex_A[i]]++;
	for (int i=nbClasses-1; i>=0; i--) probasClasses[i]=(double)repartitions[i]/nbLines_A;
	free(repartitions);

	// calcul des moyennes puis matrices de covariances
	means=(double**)malloc(nbClasses*sizeof(double*));
	cov=(double**)malloc(nbClasses*sizeof(double*));
	repartitions=(int*)malloc(nbSymbols*sizeof(int));
	int count,sumRep;
	for (int i=0; i<nbClasses; i++) {
		means[i]=(double*)calloc(nbAtt,sizeof(double));
		distribSymb.push_back((double**)malloc(attSymbolic_A.size()*sizeof(double)));
		for (int j=nbAtt-1; j>=0; j--) {
			count=0;
			switch (typesAttributs[j]) {
				case 's':
					distribSymb[i][tabIndex[j]]=(double*)malloc(nbSymbols*sizeof(double));
					for (int k=nbSymbols-1; k>=0; k--) repartitions[k]=0;
					sumRep=0;
					for (int k=nbLines_A-1; k>=0; k--) {
						if (!attSymbolic_A[tabIndex[j]][k].miss && classes_A[k]==initData[i].str) {
							repartitions[indexOfSymbol_A[tabIndex[j]][k]]++;
							sumRep++;
						}
					}
					if (sumRep>0) {
						for (int k=nbSymbols-1; k>=0; k--)
							distribSymb[i][tabIndex[j]][k]=(double)repartitions[k]/sumRep;
					}
					break;
				case 'c':
					for (int k=nbLines_A-1; k>=0; k--) {
						if (!attContinue_A[tabIndex[j]][k].miss && classes_A[k]==initData[i].str) {
							means[i][j]+=new_attContinue_A[tabIndex[j]][k];
							count++;
						}
					}
					break;
				case 'd':
					for (int k=nbLines_A-1; k>=0; k--) {
						if (!attDiscrete_A[tabIndex[j]][k].miss && classes_A[k]==initData[i].str) {
							means[i][j]+=new_attDiscrete_A[tabIndex[j]][k];
							count++;
						}
					}
					break;
				default: break; // ce cas inclue les identifiants 'i'
			}
			if (count>0) means[i][j]/=count;
		}

		cov[i]=(double*)calloc(nbAtt*nbAtt,sizeof(double));
		for (int j=nbAtt-1; j>=0; j--) {
			// calcul des termes de covariance entre att_j et att_k:
			switch (typesAttributs[j]) {
				case 's':
					for (int k=j; k>=0; k--) {
						count=0;
						switch (typesAttributs[k]) {
							case 's':
								for (int l=nbLines_A-1; l>=0; l--) {
									if (!attSymbolic_A[tabIndex[j]][l].miss && !attSymbolic_A[tabIndex[k]][l].miss && classes_A[l]==initData[i].str) {
										cov[i][k+j*nbAtt]+=
										(1.0-distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][l]])
										*(1.0-distribSymb[i][tabIndex[k]][indexOfSymbol_A[tabIndex[k]][l]]);
										count++;
									}
								}
								break;
							case 'c':
								for (int l=nbLines_A-1; l>=0; l--) {
									if (!attSymbolic_A[tabIndex[j]][l].miss && !attContinue_A[tabIndex[k]][l].miss && classes_A[l]==initData[i].str) {
										cov[i][k+j*nbAtt]+=
										(1.0-distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][l]])
										*(new_attContinue_A[tabIndex[k]][l]-means[i][k]);
										count++;
									}
								}
								break;
							case 'd':
								for (int l=nbLines_A-1; l>=0; l--) {
									if (!attSymbolic_A[tabIndex[j]][l].miss && !attDiscrete_A[tabIndex[k]][l].miss && classes_A[l]==initData[i].str) {
										cov[i][k+j*nbAtt]+=
										(1.0-distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][l]])
										*(new_attDiscrete_A[tabIndex[k]][l]-means[i][k]);
										count++;
									}
								}
								break;
							default: break; // ce cas inclue les identifiants 'i'
						}
						if (count>0) {
							cov[i][k+j*nbAtt]/=count;
							cov[i][j+k*nbAtt]=cov[i][k+j*nbAtt];
						}
					}
					break;
				case 'c':
					for (int k=j; k>=0; k--) {
						count=0;
						switch (typesAttributs[k]) {
							case 's':
								for (int l=nbLines_A-1; l>=0; l--) {
									if (!attContinue_A[tabIndex[j]][l].miss && !attSymbolic_A[tabIndex[k]][l].miss && classes_A[l]==initData[i].str) {
										cov[i][k+j*nbAtt]+=
										(new_attContinue_A[tabIndex[j]][l]-means[i][j])
										*(1.0-distribSymb[i][tabIndex[k]][indexOfSymbol_A[tabIndex[k]][l]]);
										count++;
									}
								}
								break;
							case 'c':
								for (int l=nbLines_A-1; l>=0; l--) {
									if (!attContinue_A[tabIndex[j]][l].miss && !attContinue_A[tabIndex[k]][l].miss && classes_A[l]==initData[i].str) {
										cov[i][k+j*nbAtt]+=
										(new_attContinue_A[tabIndex[j]][l]-means[i][j])*(new_attContinue_A[tabIndex[k]][l]-means[i][k]);
										count++;
									}
								}
								break;
							case 'd':
								for (int l=nbLines_A-1; l>=0; l--) {
									if (!attContinue_A[tabIndex[j]][l].miss && !attDiscrete_A[tabIndex[k]][l].miss && classes_A[l]==initData[i].str) {
										cov[i][k+j*nbAtt]+=
										(new_attContinue_A[tabIndex[j]][l]-means[i][j])*(new_attDiscrete_A[tabIndex[k]][l]-means[i][k]);
										count++;
									}
								}
								break;
							default: break; // ce cas inclue les identifiants 'i'
						}
						if (count>0) {
							cov[i][k+j*nbAtt]/=count;
							cov[i][j+k*nbAtt]=cov[i][k+j*nbAtt];
						}
					}
					break;
				case 'd':
					for (int k=j; k>=0; k--) {
						count=0;
						switch (typesAttributs[k]) {
							case 's':
								for (int l=nbLines_A-1; l>=0; l--) {
									if (!attDiscrete_A[tabIndex[j]][l].miss && !attSymbolic_A[tabIndex[k]][l].miss && classes_A[l]==initData[i].str) {
										cov[i][k+j*nbAtt]+=
										(new_attDiscrete_A[tabIndex[j]][l]-means[i][j])
										*(1.0-distribSymb[i][tabIndex[k]][indexOfSymbol_A[tabIndex[k]][l]]);
										count++;
									}
								}
								break;
							case 'c':
								for (int l=nbLines_A-1; l>=0; l--) {
									if (!attDiscrete_A[tabIndex[j]][l].miss && !attContinue_A[tabIndex[k]][l].miss && classes_A[l]==initData[i].str) {
										cov[i][k+j*nbAtt]+=
										(new_attDiscrete_A[tabIndex[j]][l]-means[i][j])*(new_attContinue_A[tabIndex[k]][l]-means[i][k]);
										count++;
									}
								}
								break;
							case 'd':
								for (int l=nbLines_A-1; l>=0; l--) {
									if (!attDiscrete_A[tabIndex[j]][l].miss && !attDiscrete_A[tabIndex[k]][l].miss && classes_A[l]==initData[i].str) {
										cov[i][k+j*nbAtt]+=
										(new_attDiscrete_A[tabIndex[j]][l]-means[i][j])*(new_attDiscrete_A[tabIndex[k]][l]-means[i][k]);
										count++;
									}
								}
								break;
							default: break; // ce cas inclue les identifiants 'i'
						}
						if (count>0) {
							cov[i][k+j*nbAtt]/=count;
							cov[i][j+k*nbAtt]=cov[i][k+j*nbAtt];
						}
					}
					break;
				default: break; // cas de l'identifiant 'i'
			}
		}
	}

	// on initialise aussi les determinants et matrices inverses pour les tests:
	determinants=(double*)malloc(nbClasses*sizeof(double));
	inv_cov=(double**)malloc(nbClasses*sizeof(double*));
	for (int i=nbClasses-1; i>=0; i--)
		inv_cov[i]=(double*)malloc(nbAtt*nbAtt*sizeof(double));
	initClassify(nbClasses);
}

// classifie l'exemple de la ligne numLine, supposant l'appel a la fonction initialize() realise
string& ClassifGauss::classify(int numLine,bool testData) {
	double proba,bestProba=-1.0;
	int bestIndex=-1;
	for (int i=nbClasses-1; i>=0; i--) {
		// on calcule les probabilites de chacune des classes, sachant la donnee
		proba=probasClasses[i];
		if (determinants[i]>0.0) proba*=(1.0/sqrt(determinants[i]))*exp(-0.5*normeCarree(numLine,i,testData));
		if (proba>bestProba) {
			bestProba=proba;
			bestIndex=i;
		}
	}

	return initData[bestIndex].str;
}

// affichage des parametres dans un fichier
// remarque: les resultats de calculs avec valeurs inutiles (identifiants) sont simplement a zero
void ClassifGauss::printToFile() {
	ofstream outFile("results/paramGauss.txt");
	bool meansPrinted;
	outFile << "/// probabilites des classes: ///" << endl << endl;
	for (int i=0; i<nbClasses; i++) outFile << initData[i].str << "  " << probasClasses[i] << endl;
	outFile << endl << "/// moyennes et matrices de covariance des classes: ///" << endl << endl;
	for (int i=0; i<nbClasses; i++) {
		meansPrinted=false;
		outFile << "*** " << initData[i].str << " ***" << endl << endl;
		for (int j=0; j<nbAtt; j++) {
			if (!meansPrinted) {
				outFile << "MEAN=" << endl << "   ";
				meansPrinted=true;
			}
			outFile << fixed << means[i][j] << " ";
		}
		if (meansPrinted) {
			outFile << endl;
			outFile << "COV_MATRIX=" << endl;
			for (int j=0; j<nbAtt; j++) {
				outFile << "  ";
				for (int k=0; k<nbAtt; k++) {
					if (cov[i][k+j*nbAtt]>=0.0) outFile << " ";
					outFile << fixed << cov[i][k+j*nbAtt] << " ";
				}
				outFile << endl;
			}
		}
		if (!attSymbolic_A.empty()) {
			if (meansPrinted) outFile << endl;
			outFile << "les probabilites des symboles respectifs suivants: " << endl;
			for (int j=nbSymbols-1; j>=0; j--) outFile << symbols[j] << "  ";
			outFile << endl << "sont, par attribut:" << endl;
		}
		for (int j=0; j<(int)attSymbolic_A.size(); j++) {
			outFile << (j+1) << ":" << endl;
			for (int k=nbSymbols-1; k>=0; k--) outFile << distribSymb[i][j][k] << " ";
			outFile << endl;
		}
		outFile << endl;
		if (!attSymbolic_A.empty()) outFile << endl;
	}
}
