#include "ClassifGaussMixt.h"

// initialise le classifieur
void ClassifGaussMixt::initialize() {
	// il faut determiner le nombre de modes (=gaussiennes) "pour chaque classe"
	// pour cela on utilise simplement l'algorithme k-means, modifie pour que k puisse augmenter
	// Ceci permet l'initialisation des P(j), mu_j (vecteurs moyennes) et Sigma_j (matrices de covariance);
	ad_k_means();

	// on initialise aussi les determinants et matrices inverses pour les tests:
	determinants=(double*)malloc(nbModes*sizeof(double));
	inv_cov=(double**)malloc(nbModes*sizeof(double*));
	for (int i=nbModes-1; i>=0; i--)
		inv_cov[i]=(double*)malloc(nbAtt*nbAtt*sizeof(double));
	initClassify(nbModes);

	// enfin, l'algorithme EM pour raffiner les resultats obtenus par ad_k-means
	exp_max();
}

// "adaptive k-means" en typesAttributs.size() dimensions (moins les identifiants)
/* extension future:
 * trouver les dimensions pertinentes via des tests statistiques,
 * et faire les calculs pour les medoides/centroides seulement sur
 * ces dimensions, un peu comme dans HPStream */
void ClassifGaussMixt::ad_k_means() {
	// on travaille avec les numeros de lignes (suffisants pour reconstituer un exemple);
	// on demarre avec k=nbClasses clusters, puis on augmente k tant que l'on ameliore le clustering
	// on initialise les k premiers centroides avec ceux des classes correspondantes:
	vector<double*> centroids;
	int count,sumRep;
	for (int i=0; i<nbClasses; i++) {
		centroids.push_back((double*)calloc(nbAtt,sizeof(double)));
		distribSymb.push_back((double**)malloc(attSymbolic_A.size()*sizeof(double*)));
		for (int j=nbAtt-1; j>=0; j--) {
			count=0;
			switch (typesAttributs[j]) {
				case 's':
					distribSymb[i][tabIndex[j]]=(double*)calloc(nbSymbols,sizeof(double));
					sumRep=0;
					for (int l=nbLines_A-1; l>=0; l--) {
						if (!attSymbolic_A[tabIndex[j]][l].miss && classes_A[l]==initData[i].str) {
							distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][l]]+=1.0;
							sumRep++;
						}
					}
					if (sumRep>0) {
						for (int l=nbSymbols-1; l>=0; l--)
							distribSymb[i][tabIndex[j]][l]/=sumRep;
					}
					break;
				case 'c':
					for (int l=nbLines_A-1; l>=0; l--) {
						if (!attContinue_A[tabIndex[j]][l].miss && classes_A[l]==initData[i].str) {
							centroids[i][j]+=new_attContinue_A[tabIndex[j]][l];
							count++;
						}
					}
					break;
				case 'd':
					for (int l=nbLines_A-1; l>=0; l--) {
						if (!attDiscrete_A[tabIndex[j]][l].miss && classes_A[l]==initData[i].str) {
							centroids[i][j]+=new_attDiscrete_A[tabIndex[j]][l];
							count++;
						}
					}
					break;
				default: break; // cas de 'i'
			}
			if (count>0) centroids[i][j]/=count;
		}
	}

	// on en deduit alors k medoides en calculant les distances des points aux centroides juste trouves;
	// ces k medoides ne seront pas forcement homogenes, suivant la separation des donnees;
	vector<list<int> > medoids; // medoides representant les modes
	list<int> emptyList;
	for (int i=nbClasses-1; i>=0; i--) medoids.push_back(emptyList);
	double dist,min_dist;
	int bestIndex;
	for (int i=nbLines_A-1; i>=0; i--) {
		min_dist=MAX_DOUBLE;
		bestIndex=-1;
		for (int j=nbClasses-1; j>=0; j--) {
			// on calcule la distance de l'exemple i au centroide j
			dist=0.0;
			count=0;
			for (int l=nbAtt-1; l>=0; l--) {
				switch (typesAttributs[l]) {
					case 's':
						if (!attSymbolic_A[tabIndex[l]][i].miss) {
							dist+=(1.0-distribSymb[j][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][i]]);
							count++;
						}
						break;
					case 'c':
						if (!attContinue_A[tabIndex[l]][i].miss) {
							dist+=fabs(new_attContinue_A[tabIndex[l]][i]-centroids[j][l]);
							count++;
						}
						break;
					case 'd':
						if (!attDiscrete_A[tabIndex[l]][i].miss) {
							dist+=fabs(new_attDiscrete_A[tabIndex[l]][i]-centroids[j][l]);
							count++;
						}
						break;
					default: break; // cas de 'i'
				}
			}
			if (count>0) dist/=count;
			if (dist<min_dist) {
				min_dist=dist;
				bestIndex=j;
			}
		}
		medoids[bestIndex].push_back(i);
	}

	// on itere le processus jusqu'a epsilon-stabilisation:
	vector<double**> new_distribSymb(nbClasses);
	vector<double*> new_centroids(nbClasses);
	for (int i=nbClasses-1; i>=0; i--) {
		new_distribSymb[i]=(double**)malloc(attSymbolic_A.size()*sizeof(double*));
		for (int j=attSymbolic_A.size()-1; j>=0; j--) new_distribSymb[i][j]=(double*)malloc(nbSymbols*sizeof(double));
		new_centroids[i]=(double*)malloc(nbAtt*sizeof(double));
	}
	double aux,deltaMax,old_deltaMax=MAX_DOUBLE;
	list<int>::iterator curInt,movedInt;
	while (true) {
		// on calcule les nouveaux centroides, que l'on initialise a zero
		for (int i=nbClasses-1; i>=0; i--) {
			for (int j=nbAtt-1; j>=0; j--) {
				new_centroids[i][j]=0.0;
				count=0;
				switch (typesAttributs[j]) {
					case 's':
						for (int l=nbSymbols-1; l>=0; l--) new_distribSymb[i][tabIndex[j]][l]=0.0;
						sumRep=0;
						for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
							if (!attSymbolic_A[tabIndex[j]][*curInt].miss) {
								new_distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][*curInt]]+=1.0;
								sumRep++;
							}
						}
						if (sumRep>0) {
							for (int l=nbSymbols-1; l>=0; l--)
								new_distribSymb[i][tabIndex[j]][l]/=sumRep;
						}
						break;
					case 'c':
						for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
							if (!attContinue_A[tabIndex[j]][*curInt].miss) {
								new_centroids[i][j]+=new_attContinue_A[tabIndex[j]][*curInt];
								count++;
							}
						}
						break;
					case 'd':
						for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
							if (!attDiscrete_A[tabIndex[j]][*curInt].miss) {
								new_centroids[i][j]+=new_attDiscrete_A[tabIndex[j]][*curInt];
								count++;
							}
						}
						break;
					default: break; // cas de 'i'
				}
				if (count>0) new_centroids[i][j]/=count;
			}
		}

		// on cherche l'ecart maximal entre un ancien et nouveau centroide
		deltaMax=-1.0;
		for (int i=nbClasses-1; i>=0; i--) {
			dist=0.0;
			for (int j=nbAtt-1; j>=0; j--) {
				if (typesAttributs[j]=='s') {
					aux=0.0;
					for (int l=nbSymbols-1; l>=0; l--)
						aux+=fabs(new_distribSymb[i][tabIndex[j]][l]-distribSymb[i][tabIndex[j]][l]);
					aux/=nbSymbols;
					dist+=aux;
				}
				else if (typesAttributs[j]!='i') {
					dist+=fabs(new_centroids[i][j]-centroids[i][j]);
				}
			}
			dist/=nbRealAtt;
			if (dist>deltaMax) deltaMax=dist;
		}
		// si cet ecart est assez proche de la derniere valeur de deltaMax, on sort
		if (fabs(old_deltaMax-deltaMax)<EPSILON) break;
		old_deltaMax=deltaMax;

		// sinon, on passe a l'iteration suivante, centroids:=new_centroids
		for (int i=nbClasses-1; i>=0; i--) {
			for (int j=attSymbolic_A.size()-1; j>=0; j--) {
				for (int l=nbSymbols-1; l>=0; l--) distribSymb[i][j][l]=new_distribSymb[i][j][l];
			}
			for (int j=nbAtt-1; j>=0; j--) centroids[i][j]=new_centroids[i][j];
		}
		// on recalcule alors les medoides
		for (int i=nbClasses-1; i>=0; i--) {
			curInt=medoids[i].begin();
			while (curInt!=medoids[i].end()) {
				min_dist=MAX_DOUBLE;
				bestIndex=-1;
				for (int j=nbClasses-1; j>=0; j--) {
					// on calcule la distance de l'exemple curInt au centroide j
					dist=0.0;
					count=0;
					for (int l=nbAtt-1; l>=0; l--) {
						switch (typesAttributs[l]) {
							case 's':
								if (!attSymbolic_A[tabIndex[l]][*curInt].miss) {
									dist+=(1.0-distribSymb[j][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][*curInt]]);
									count++;
								}
								break;
							case 'c':
								if (!attContinue_A[tabIndex[l]][*curInt].miss) {
									dist+=fabs(new_attContinue_A[tabIndex[l]][*curInt]-centroids[j][l]);
									count++;
								}
								break;
							case 'd':
								if (!attDiscrete_A[tabIndex[l]][*curInt].miss) {
									dist+=fabs(new_attDiscrete_A[tabIndex[l]][*curInt]-centroids[j][l]);
									count++;
								}
								break;
							default: break; // cas de 'i'
						}
					}
					if (count>0) dist/=count;
					if (dist<min_dist) {
						min_dist=dist;
						bestIndex=j;
					}
				}
				// replacement dans le bon medoide si bestIndex!=i:
				if (bestIndex!=i) {
					medoids[bestIndex].push_back(*curInt);
					movedInt=curInt;
					curInt++;
					medoids[i].erase(movedInt);
				}
				else curInt++;
			}
		}
	}

	nbModes=nbClasses;
	int index,index_m;
	double sum,stdDev,max_stdDev,stdDev_m;
	bool inFirst,go_on=true;
	vector<IndexAndProba> triProbSymb(nbSymbols);int yyy=0;
	// on continue d'augmenter nbModes tant que la separation des medoides reste correcte
	// et que ceux-ci ont assez d'elements.
	while (yyy++<50) {
		// on calcule l'ecart-type maximal unidimensionnel par medoide, que l'on compare avec
		// la distance au plus proche voisin le long de cette meme dimension:
		// si celui-ci est plus grand pour au moins un medoide,
		// alors on continue d'ajouter des centroides. Sinon on s'arrete sans rien ajouter.
		// On s'arrete aussi si un medoide est trop petit, signifiant qu'il y a trop de modes.
		// Si necessaire, on place deux centroides assez eloignes sur l'axe de l'ecart-type maximal,
		// puis on itere jusqu'a stabilisation avant de refaire les tests de continuation.
		max_stdDev=-1.0;
		bestIndex=-1;
		index=-1;
		go_on=false;
		for (int i=nbModes-1; i>=0; i--) {
			if (medoids[i].size()<(unsigned int)floor((SIZE_FACTOR*nbLines_A)/nbModes)) {
				go_on=false;
				break;
			}
			stdDev_m=-1.0;
			index_m=-1;
			for (int j=nbAtt-1; j>=0; j--) {
				stdDev=0.0;
				count=0;
				switch (typesAttributs[j]) {
					case 's':
						for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
							if (!attSymbolic_A[tabIndex[j]][*curInt].miss) {
								stdDev+=square(1.0-distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][*curInt]]);
								count++;
							}
						}
						break;
					case 'c':
						for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
							if (!attContinue_A[tabIndex[j]][*curInt].miss) {
								stdDev+=square(new_attContinue_A[tabIndex[j]][*curInt]-centroids[i][j]);
								count++;
							}
						}
						break;
					case 'd':
						for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
							if (!attDiscrete_A[tabIndex[j]][*curInt].miss) {
								stdDev+=square(new_attDiscrete_A[tabIndex[j]][*curInt]-centroids[i][j]);
								count++;
							}
						}
						break;
					default: break; // cas de 'i'
				}
				if (count>0) stdDev=sqrt(stdDev/count);
				if (stdDev>stdDev_m) {
					stdDev_m=stdDev;
					index_m=j;
				}
			}

			// maintenant on calcule la distance du centroide i a chaque autre centroide,
			// le long de la bonne dimension
			for (int j=nbModes-1; j>=0; j--) {
				if (j!=i) {
					dist=0.0;
					if (typesAttributs[index_m]=='s') {
						for (int m=nbSymbols-1; m>=0; m--)
							dist+=fabs(distribSymb[j][tabIndex[index_m]][m]-distribSymb[i][tabIndex[index_m]][m]);
						dist/=nbSymbols;
					}
					else dist=fabs(centroids[j][index_m]-centroids[i][index_m]);
					if (stdDev_m>CLUSTERING_THRESHOLD*nbModes*dist && stdDev_m>max_stdDev) {
						go_on=true;
						bestIndex=i;
						index=index_m;
						max_stdDev=stdDev_m;
					}
				}
			}
		}
		if (!go_on) {
			// cas ou tous les clusters sont juges suffisamment purs;
			// on n'ajoute pas de centroide, et on s'arrete:
			break;
		}

		// on constitue la nouvelle liste des medoides et des centroides,
		distribSymb.push_back((double**)malloc(attSymbolic_A.size()*sizeof(double*)));
		for (int i=attSymbolic_A.size()-1; i>=0; i--)
			distribSymb.back()[i]=(double*)malloc(nbSymbols*sizeof(double));
		centroids.push_back((double*)malloc(nbAtt*sizeof(double)));
		if (typesAttributs[index]=='s') {
			// dans le cas particulier ou l'ecart-type max est pour une valeur symbolique,
			// il faut diviser le cluster en 2 parties aux distributions differentes:
			/* NOTE: on rencontre ici un probleme NP-complet:
			* trouver la meilleure separation d'un ensemble de rels
			* telle que les sommes des elements dans chaque partie soit
			* "a peu pres" egale.
			* On l'approxime avec l'heuristique qui consiste a toujours choisir
			* le plus petit element pour le prochain ajout */
			for (int i=nbSymbols-1; i>=0; i--) {
				triProbSymb[i].index=i;
				triProbSymb[i].proba=distribSymb[bestIndex][tabIndex[index]][i];
			}
			sort(triProbSymb.begin(),triProbSymb.end());
			sum=0.0;
			int curInd=0;
			while (sum<0.5) {
				sum+=triProbSymb[curInd].proba;
				curInd++;
			}

			// initialisation des premiers groupes de symboles:
			for (int i=nbAtt-1; i>=0; i--) {
				if (i!=index && typesAttributs[i]=='s') {
					for (int j=nbSymbols-1; j>=0; j--) distribSymb[bestIndex][tabIndex[i]][j]=0.0;
					sumRep=0;
					for (curInt=medoids[bestIndex].begin(); curInt!=medoids[bestIndex].end(); curInt++) {
						if (!attSymbolic_A[tabIndex[i]][*curInt].miss) {
							inFirst=false;
							for (int j=curInd-1; j>=0; j--) {
								if (triProbSymb[j].index==indexOfSymbol_A[tabIndex[i]][*curInt]) {
									inFirst=true;
									break;
								}
							}
							if (inFirst) {
								distribSymb[bestIndex][tabIndex[i]][indexOfSymbol_A[tabIndex[i]][*curInt]]+=1.0;
								sumRep++;
							}
						}
					}
					if (sumRep>0) {
						for (int j=nbSymbols-1; j>=0; j--)
							distribSymb[bestIndex][tabIndex[i]][j]/=sumRep;
					}
				}
				else {
					for (int j=nbSymbols-1; j>=0; j--) distribSymb[bestIndex][tabIndex[index]][j]=0.0;
					sumRep=0;
					for (curInt=medoids[bestIndex].begin(); curInt!=medoids[bestIndex].end(); curInt++) {
						if (!attSymbolic_A[tabIndex[index]][*curInt].miss) {
							inFirst=false;
							for (int j=curInd-1; j>=0; j--) {
								if (triProbSymb[j].index==indexOfSymbol_A[tabIndex[index]][*curInt]) {
									inFirst=true;
									break;
								}
							}
							if (inFirst) {
								distribSymb[bestIndex][tabIndex[index]][indexOfSymbol_A[tabIndex[index]][*curInt]]+=1.0;
								sumRep++;
							}
						}
					}
					if (sumRep>0) {
						for (int j=nbSymbols-1; j>=0; j--)
							distribSymb[bestIndex][tabIndex[index]][j]/=sumRep;
					}
				}
			}

			// initialisation des seconds groupes de symboles:
			for (int i=nbAtt-1; i>=0; i--) {
				if (i!=index && typesAttributs[i]=='s') {
					for (int j=nbSymbols-1; j>=0; j--) distribSymb.back()[tabIndex[i]][j]=0.0;
					sumRep=0;
					for (curInt=medoids[bestIndex].begin(); curInt!=medoids[bestIndex].end(); curInt++) {
						if (!attSymbolic_A[tabIndex[i]][*curInt].miss) {
							inFirst=false;
							for (int j=curInd-1; j>=0; j--) {
								if (triProbSymb[j].index==indexOfSymbol_A[tabIndex[i]][*curInt]) {
									inFirst=true;
									break;
								}
							}
							if (inFirst) {
								distribSymb.back()[tabIndex[i]][indexOfSymbol_A[tabIndex[i]][*curInt]]+=1.0;
								sumRep++;
							}
						}
					}
					if (sumRep>0) {
						for (int j=nbSymbols-1; j>=0; j--)
							distribSymb.back()[tabIndex[i]][j]/=sumRep;
					}
				}
				else {
					for (int j=nbSymbols-1; j>=0; j--) distribSymb.back()[tabIndex[index]][j]=0.0;
					sumRep=0;
					for (curInt=medoids[bestIndex].begin(); curInt!=medoids[bestIndex].end(); curInt++) {
						if (!attSymbolic_A[tabIndex[index]][*curInt].miss) {
							inFirst=false;
							for (int j=curInd-1; j>=0; j--) {
								if (triProbSymb[j].index==indexOfSymbol_A[tabIndex[index]][*curInt]) {
									inFirst=true;
									break;
								}
							}
							if (inFirst) {
								distribSymb.back()[tabIndex[index]][indexOfSymbol_A[tabIndex[index]][*curInt]]+=1.0;
								sumRep++;
							}
						}
					}
					if (sumRep>0) {
						for (int j=nbSymbols-1; j>=0; j--)
							distribSymb.back()[tabIndex[index]][j]/=sumRep;
					}
				}
			}

			// enfin, on recopie simplement les valeurs numeriques centroids[bestIndex]:
			for (int i=nbAtt-1; i>=0; i--) centroids.back()[i]=centroids[bestIndex][i];
		}
		else {
			for (int i=nbAtt-1; i>=0; i--) {
				if (i!=index) centroids.back()[i]=centroids[bestIndex][i];
				else {
					centroids.back()[index]=centroids[bestIndex][index]+PLACE_CENTROID*max_stdDev;
					centroids[bestIndex][index]=centroids[bestIndex][index]-PLACE_CENTROID*max_stdDev;
				}
			}

			// ensuite, on recopie simplement les valeurs symboliques distribSymb[bestIndex]:
			for (int i=attSymbolic_A.size()-1; i>=0; i--) {
				for (int j=nbSymbols-1; j>=0; j--) distribSymb.back()[i][j]=distribSymb[bestIndex][i][j];
			}
		}

		// on en deduit les nouveaux medoides
		medoids.push_back(emptyList);
		for (int i=nbModes-1; i>=0; i--) {
			curInt=medoids[i].begin();
			while (curInt!=medoids[i].end()) {
				min_dist=MAX_DOUBLE;
				bestIndex=-1;
				for (int j=nbModes; j>=0; j--) {
					// on calcule la distance de l'exemple curInt au centroide j
					dist=0.0;
					count=0;
					for (int l=nbAtt-1; l>=0; l--) {
						switch (typesAttributs[l]) {
							case 's':
								if (!attSymbolic_A[tabIndex[l]][*curInt].miss) {
									dist+=(1.0-distribSymb[j][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][*curInt]]);
									count++;
								}
								break;
							case 'c':
								if (!attContinue_A[tabIndex[l]][*curInt].miss) {
									dist+=fabs(new_attContinue_A[tabIndex[l]][*curInt]-centroids[j][l]);
									count++;
								}
								break;
							case 'd':
								if (!attDiscrete_A[tabIndex[l]][*curInt].miss) {
									dist+=fabs(new_attDiscrete_A[tabIndex[l]][*curInt]-centroids[j][l]);
									count++;
								}
								break;
							default: break; // cas de 'i'
						}
					}
					if (count>0) dist/=count;
					if (dist<min_dist) {
						min_dist=dist;
						bestIndex=j;
					}
				}
				// placement dans le bon medoide, s'il n'y est deja:
				if (bestIndex!=i) {
					medoids[bestIndex].push_back(*curInt);
					movedInt=curInt;
					curInt++;
					medoids[i].erase(movedInt);
				}
				else curInt++;
			}
		}
		new_distribSymb.push_back((double**)malloc(attSymbolic_A.size()*sizeof(double*)));
		for (int i=attSymbolic_A.size()-1; i>=0; i--) new_distribSymb.back()[i]=(double*)malloc(nbSymbols*sizeof(double));
		new_centroids.push_back((double*)malloc(nbAtt*sizeof(double)));
		nbModes++;

		// iterations pour affiner les nouveaux centroides/medoides
		old_deltaMax=MAX_DOUBLE;
		while (true) {
			// on calcule les nouveaux centroides, que l'on initialise a zero
			for (int i=nbModes-1; i>=0; i--) {
				for (int j=nbAtt-1; j>=0; j--) {
					new_centroids[i][j]=0.0;
					count=0;
					switch (typesAttributs[j]) {
						case 's':
							for (int l=nbSymbols-1; l>=0; l--) new_distribSymb[i][tabIndex[j]][l]=0.0;
							sumRep=0;
							for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
								if (!attSymbolic_A[tabIndex[j]][*curInt].miss) {
									new_distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][*curInt]]+=1.0;
									sumRep++;
								}
							}
							if (sumRep>0) {
								for (int l=nbSymbols-1; l>=0; l--)
									new_distribSymb[i][tabIndex[j]][l]/=sumRep;
							}
							break;
						case 'c':
							for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
								if (!attContinue_A[tabIndex[j]][*curInt].miss) {
									new_centroids[i][j]+=new_attContinue_A[tabIndex[j]][*curInt];
									count++;
								}
							}
							break;
						case 'd':
							for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
								if (!attDiscrete_A[tabIndex[j]][*curInt].miss) {
									new_centroids[i][j]+=new_attDiscrete_A[tabIndex[j]][*curInt];
									count++;
								}
							}
							break;
						default: break; // cas de 'i'
					}
					if (count>0) new_centroids[i][j]/=count;
				}
			}

			// on cherche l'ecart maximal entre un "ancien" et nouveau centroide
			deltaMax=-1.0;
			for (int i=nbModes-1; i>=0; i--) {
				dist=0.0;
				for (int j=nbAtt-1; j>=0; j--) {
					if (typesAttributs[j]=='s') {
						aux=0.0;
						for (int l=nbSymbols-1; l>=0; l--)
							aux+=fabs(new_distribSymb[i][tabIndex[j]][l]-distribSymb[i][tabIndex[j]][l]);
						aux/=nbSymbols;
						dist+=aux;
					}
					else if (typesAttributs[j]!='i') {
						dist+=fabs(new_centroids[i][j]-centroids[i][j]);
					}
				}
				dist/=nbRealAtt;
				if (dist>deltaMax) deltaMax=dist;
			}
			// si cet ecart est assez proche de la derniere valeur de deltaMax, on sort
			if (fabs(old_deltaMax-deltaMax)<EPSILON) break;
			old_deltaMax=deltaMax;

			// sinon, on passe a l'iteration suivante, centroids:=new_centroids
			for (int i=nbModes-1; i>=0; i--) {
				for (int j=attSymbolic_A.size()-1; j>=0; j--) {
					for (int l=nbSymbols-1; l>=0; l--) distribSymb[i][j][l]=new_distribSymb[i][j][l];
				}
				for (int j=nbAtt-1; j>=0; j--) centroids[i][j]=new_centroids[i][j];
			}
			// on recalcule alors les medoides
			for (int i=nbModes-1; i>=0; i--) {
				curInt=medoids[i].begin();
				while (curInt!=medoids[i].end()) {
					min_dist=MAX_DOUBLE;
					bestIndex=-1;
					for (int j=nbModes-1; j>=0; j--) {
						// calcul de la distance de l'exemple curInt au centroide j
						dist=0.0;
						count=0;
						for (int l=nbAtt-1; l>=0; l--) {
							switch (typesAttributs[l]) {
								case 's':
									if (!attSymbolic_A[tabIndex[l]][*curInt].miss) {
										dist+=(1.0-distribSymb[j][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][*curInt]]);
										count++;
									}
									break;
								case 'c':
									if (!attContinue_A[tabIndex[l]][*curInt].miss) {
										dist+=fabs(new_attContinue_A[tabIndex[l]][*curInt]-centroids[j][l]);
										count++;
									}
									break;
								case 'd':
									if (!attDiscrete_A[tabIndex[l]][*curInt].miss) {
										dist+=fabs(new_attDiscrete_A[tabIndex[l]][*curInt]-centroids[j][l]);
										count++;
									}
									break;
								default: break; // cas de 'i'
							}
						}
						if (count>0) dist/=count;
						if (dist<min_dist) {
							min_dist=dist;
							bestIndex=j;
						}
					}
					// replacement dans le bon medoide si bestIndex!=i:
					if (bestIndex!=i) {
						medoids[bestIndex].push_back(*curInt);
						movedInt=curInt;
						curInt++;
						medoids[i].erase(movedInt);
					}
					else curInt++;
				}
			}
		}
	}

	// on libere l'espace memoire des centroides:
	for (int i=centroids.size()-1; i>=0; i--) free(centroids[i]);
	for (int i=new_centroids.size()-1; i>=0; i--) free(new_centroids[i]);

	// enfin, on initialise probasClassModes, means et cov avec les valeurs obtenues
	probasClassModes=new ClassAndProba[nbModes];
	means=(double**)malloc(nbModes*sizeof(double*));
	cov=(double**)malloc(nbModes*sizeof(double*));
	for (int i=nbModes-1; i>=0; i--) {
		probasClassModes[i].proba=(double)medoids[i].size()/nbLines_A;

		means[i]=(double*)malloc(nbAtt*sizeof(double));
		for (int j=nbAtt-1; j>=0; j--) {
			means[i][j]=0.0;
			count=0;
			switch (typesAttributs[j]) {
				case 's':
					for (int l=nbSymbols-1; l>=0; l--) distribSymb[i][tabIndex[j]][l]=0.0;
					sumRep=0;
					for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
						if (!attSymbolic_A[tabIndex[j]][*curInt].miss) {
							distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][*curInt]]+=1.0;
							sumRep++;
						}
					}
					if (sumRep>0) {
						for (int l=nbSymbols-1; l>=0; l--)
							distribSymb[i][tabIndex[j]][*curInt]/=sumRep;
					}
					break;
				case 'c':
					for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
						if (!attContinue_A[tabIndex[j]][*curInt].miss) {
							means[i][j]+=new_attContinue_A[tabIndex[j]][*curInt];
							count++;
						}
					}
					break;
				case 'd':
					for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
						if (!attDiscrete_A[tabIndex[j]][*curInt].miss) {
							means[i][j]+=new_attDiscrete_A[tabIndex[j]][*curInt];
							count++;
						}
					}
					break;
				default: break; // ce cas inclue les identifiants 'i'
			}
			if (count>0) means[i][j]/=count;
		}

		cov[i]=(double*)calloc(nbAtt*nbAtt,sizeof(double));
		for (int j=nbAtt-1; j>=0; j--){
			// calcul des termes de covariance entre att_j et att_k:
			switch (typesAttributs[j]) {
				case 's':
					for (int l=j; l>=0; l--) {
						count=0;
						switch (typesAttributs[l]) {
							case 's':
								for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
									if (!attSymbolic_A[tabIndex[j]][*curInt].miss && !attSymbolic_A[tabIndex[l]][*curInt].miss) {
										cov[i][l+j*nbAtt]+=
										(1.0-distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][*curInt]])
										*(1.0-distribSymb[i][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][*curInt]]);
										count++;
									}
								}
								break;
							case 'c':
								for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
									if (!attSymbolic_A[tabIndex[j]][*curInt].miss && !attContinue_A[tabIndex[l]][*curInt].miss) {
										cov[i][l+j*nbAtt]+=
										(1.0-distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][*curInt]])
										*(new_attContinue_A[tabIndex[l]][*curInt]-means[i][l]);
										count++;
									}
								}
								break;
							case 'd':
								for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
									if (!attSymbolic_A[tabIndex[j]][*curInt].miss && !attDiscrete_A[tabIndex[l]][*curInt].miss) {
										cov[i][l+j*nbAtt]+=
										(1.0-distribSymb[i][tabIndex[j]][indexOfSymbol_A[tabIndex[j]][*curInt]])
										*(new_attDiscrete_A[tabIndex[l]][*curInt]-means[i][l]);
										count++;
									}
								}
								break;
							default: break; // ce cas inclue les identifiants 'i'
						}
						if (count>0) {
							cov[i][l+j*nbAtt]/=count;
							cov[i][j+l*nbAtt]=cov[i][l+j*nbAtt];
						}
					}
					break;
				case 'c':
					for (int l=j; l>=0; l--) {
						count=0;
						switch (typesAttributs[l]) {
							case 's':
								for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
									if (!attContinue_A[tabIndex[j]][*curInt].miss && !attSymbolic_A[tabIndex[l]][*curInt].miss) {
										cov[i][l+j*nbAtt]+=
										(new_attContinue_A[tabIndex[j]][*curInt]-means[i][j])
										*(1.0-distribSymb[i][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][*curInt]]);
										count++;
									}
								}
								break;
							case 'c':
								for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
									if (!attContinue_A[tabIndex[j]][*curInt].miss && !attContinue_A[tabIndex[l]][*curInt].miss) {
										cov[i][l+j*nbAtt]+=
										(new_attContinue_A[tabIndex[j]][*curInt]-means[i][j])*(new_attContinue_A[tabIndex[l]][*curInt]-means[i][l]);
										count++;
									}
								}
								break;
							case 'd':
								for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
									if (!attContinue_A[tabIndex[j]][*curInt].miss && !attDiscrete_A[tabIndex[l]][*curInt].miss) {
										cov[i][l+j*nbAtt]+=
										(new_attContinue_A[tabIndex[j]][*curInt]-means[i][j])*(new_attDiscrete_A[tabIndex[l]][*curInt]-means[i][l]);
										count++;
									}
								}
								break;
							default: break; // ce cas inclue les identifiants 'i'
						}
						if (count>0) {
							cov[i][l+j*nbAtt]/=count;
							cov[i][j+l*nbAtt]=cov[i][l+j*nbAtt];
						}
					}
					break;
				case 'd':
					for (int l=j; l>=0; l--) {
						count=0;
						switch (typesAttributs[l]) {
							case 's':
								for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
									if (!attDiscrete_A[tabIndex[j]][*curInt].miss && !attSymbolic_A[tabIndex[l]][*curInt].miss) {
										cov[i][l+j*nbAtt]+=
										(new_attDiscrete_A[tabIndex[j]][*curInt]-means[i][j])
										*(1.0-distribSymb[i][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][*curInt]]);
										count++;
									}
								}
								break;
							case 'c':
								for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
									if (!attDiscrete_A[tabIndex[j]][*curInt].miss && !attContinue_A[tabIndex[l]][*curInt].miss) {
										cov[i][l+j*nbAtt]+=
										(new_attDiscrete_A[tabIndex[j]][*curInt]-means[i][j])*(new_attContinue_A[tabIndex[l]][*curInt]-means[i][l]);
										count++;
									}
								}
								break;
							case 'd':
								for (curInt=medoids[i].begin(); curInt!=medoids[i].end(); curInt++) {
									if (!attDiscrete_A[tabIndex[j]][*curInt].miss && !attDiscrete_A[tabIndex[l]][*curInt].miss) {
										cov[i][l+j*nbAtt]+=
										(new_attDiscrete_A[tabIndex[j]][*curInt]-means[i][j])*(new_attDiscrete_A[tabIndex[l]][*curInt]-means[i][l]);
										count++;
									}
								}
								break;
							default: break; // ce cas inclue les identifiants 'i'
						}
						if (count>0) {
							cov[i][l+j*nbAtt]/=count;
							cov[i][j+l*nbAtt]=cov[i][l+j*nbAtt];
						}
					}
					break;
				default: break; // cas de l'identifiant 'i'
			}
		}
	}
}

// l'algorithme EM, supposant l'initialisation precedente realisee:
// les medoides/centroides continuent d'evoluer implicitement mais k n'augmente plus
/* NOTE: EM ne peut pas s'appliquer simplement aux attributs symboliques, meme transformes;
 * on les ignore donc dans les iterations */
void ClassifGaussMixt::exp_max() {
	// on a les valeurs initiales des mu_j et Sigma_j dans means et cov
	// allocations des nouvelles valeurs:
	double* new_probasClassModes=(double*)malloc(nbModes*sizeof(double));
	double** new_means=(double**)malloc(nbModes*sizeof(double*));
	double** new_cov=(double**)malloc(nbModes*sizeof(double*));
	for (int i=nbModes-1; i>=0; i--) {
		new_means[i]=(double*)malloc(nbAtt*sizeof(double));
		new_cov[i]=(double*)malloc(nbAtt*nbAtt*sizeof(double));
	}

	double dist,proba,sumProba,sum;
	int count;
	if (!attContinue_A.empty() || !attDiscrete_A.empty()) {
		// constante de convergence, depend de la dimensionnalite:
		double EPS_EM=0.01*pow(attContinue_A.size()+attDiscrete_A.size(),1.5);
		bool go_on=false;
		while (true) {
			/* amelioration possible pour efficacite: tabuler les probabilites calculees;
			 * ca permettrait d'inverser les deux boucles, mais couteux en memoire. */
			for (int i=nbModes-1; i>=0; i--) {
				sumProba=0.0;
				count=0;
				for (int j=nbAtt-1; j>=0; j--) {
					new_means[i][j]=0.0;
					for (int l=nbAtt-1; l>=0; l--) new_cov[i][l+j*nbAtt]=0.0;
				}
				for (int j=nbLines_A-1; j>=0; j--) {
					// calcul de P(i|x)=P(i)p(x|i)/somme des P(a)p(x|a)
					proba=probasClassModes[i].proba;
					if (determinants[i]>0.0) {
						proba*=(1.0/sqrt(determinants[i]))*exp(-0.5*normeCarree(j,i,false));
						sum=0.0;
						for (int l=nbModes-1; l>=0; l--) {
							if (determinants[l]>0.0)
								sum+=(probasClassModes[l].proba/sqrt(determinants[l]))*exp(-0.5*normeCarree(j,l,false));
						}
						if (sum>0.0) proba/=sum;
					}
					// la ligne suivante pour eviter les erreurs d'approximations numeriques:
					if (!(-MAX_DOUBLE<proba && proba<MAX_DOUBLE)) continue;
					sumProba+=proba;
					count++;
	
					// calcul des nouvelles moyennes
					for (int l=nbAtt-1; l>=0; l--) {
						switch (typesAttributs[l]) {
							case 'c':
								if (!attContinue_A[tabIndex[l]][j].miss)
									new_means[i][l]+=proba*new_attContinue_A[tabIndex[l]][j];
								break;
							case 'd':
								if (!attDiscrete_A[tabIndex[l]][j].miss)
									new_means[i][l]+=proba*new_attDiscrete_A[tabIndex[l]][j];
								break;
							default: break; // cas de 's' ou 'i'
						}
					}
	
					// calcul des nouvelles matrices de covariance
					for (int l=nbAtt-1; l>=0; l--) {
						switch (typesAttributs[l]) {
							case 's':
								if (attSymbolic_A[tabIndex[l]][j].miss) break;
								for (int m=nbAtt-1; m>=0; m--) {
									switch (typesAttributs[m]) {
										case 's':
											if (!attSymbolic_A[tabIndex[m]][j].miss)
												new_cov[i][m+l*nbAtt]+=proba
													*(1.0-distribSymb[i][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][j]])
													*(1.0-distribSymb[i][tabIndex[m]][indexOfSymbol_A[tabIndex[m]][j]]);
											break;
										case 'c':
											if (!attContinue_A[tabIndex[m]][j].miss)
												new_cov[i][m+l*nbAtt]+=proba
													*(1.0-distribSymb[i][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][j]])
													*(new_attContinue_A[tabIndex[m]][j]-means[i][m]);
											break;
										case 'd':
											if (!attDiscrete_A[tabIndex[m]][j].miss)
												new_cov[i][m+l*nbAtt]+=proba
													*(1.0-distribSymb[i][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][j]])
													*(new_attDiscrete_A[tabIndex[m]][j]-means[i][m]);
											break;
										default: break; // cas de 'i'
									}
								}
								break;
							case 'c':
								if (attContinue_A[tabIndex[l]][j].miss) break;
								for (int m=nbAtt-1; m>=0; m--) {
									switch (typesAttributs[m]) {
										case 's':
											if (!attSymbolic_A[tabIndex[m]][j].miss)
												new_cov[i][m+l*nbAtt]+=proba
													*(new_attContinue_A[tabIndex[l]][j]-means[i][l])
													*(1.0-distribSymb[i][tabIndex[m]][indexOfSymbol_A[tabIndex[m]][j]]);
											break;
										case 'c':
											if (!attContinue_A[tabIndex[m]][j].miss)
												new_cov[i][m+l*nbAtt]+=proba
													*(new_attContinue_A[tabIndex[l]][j]-means[i][l])
													*(new_attContinue_A[tabIndex[m]][j]-means[i][m]);
											break;
										case 'd':
											if (!attDiscrete_A[tabIndex[m]][j].miss)
												new_cov[i][m+l*nbAtt]+=proba
													*(new_attContinue_A[tabIndex[l]][j]-means[i][l])
													*(new_attDiscrete_A[tabIndex[m]][j]-means[i][m]);
											break;
										default: break; // cas de 'i'
									}
								}
								break;
							case 'd':
								if (attDiscrete_A[tabIndex[l]][j].miss) break;
								for (int m=nbAtt-1; m>=0; m--) {
									switch (typesAttributs[m]) {
										case 's':
											if (!attSymbolic_A[tabIndex[m]][j].miss)
												new_cov[i][m+l*nbAtt]+=proba
													*(new_attDiscrete_A[tabIndex[l]][j]-means[i][l])
													*(1.0-distribSymb[i][tabIndex[m]][indexOfSymbol_A[tabIndex[m]][j]]);
											break;
										case 'c':
											if (!attContinue_A[tabIndex[m]][j].miss)
												new_cov[i][m+l*nbAtt]+=proba
													*(new_attDiscrete_A[tabIndex[l]][j]-means[i][l])
													*(new_attContinue_A[tabIndex[m]][j]-means[i][m]);
											break;
										case 'd':
											if (!attDiscrete_A[tabIndex[m]][j].miss)
												new_cov[i][m+l*nbAtt]+=proba
													*(new_attDiscrete_A[tabIndex[l]][j]-means[i][l])
													*(new_attDiscrete_A[tabIndex[m]][j]-means[i][m]);
											break;
										default: break; // cas de 'i'
									}
								}
								break;
							default: break; // cas de 'i'
						}
					}
				}
				// calcul de la nouvelle probabilite du mode:
				if (count>0) new_probasClassModes[i]=sumProba/count;
				else new_probasClassModes[i]=probasClassModes[i].proba;
				// division par la somme des probabilites conditionnelles calculees
				if (sumProba>0.0) {
					for (int j=nbAtt-1; j>=0; j--) {
						new_means[i][j]/=sumProba;
						for (int l=nbAtt-1; l>=0; l--) new_cov[i][l+j*nbAtt]/=sumProba;
					}
				}
			}
	
			// enfin, on compare means a new_means, cov a new_cov (norme infinie)
			for (int i=nbModes-1; i>=0; i--) {
				go_on=false;
				for (int j=nbAtt-1; j>=0; j--) {
					if (typesAttributs[j]=='c' || typesAttributs[j]=='d') {
						dist=fabs(new_means[i][j]-means[i][j]);
						if (dist>=EPS_EM) {
							go_on=true;
							break;
						}
					}
					for (int l=nbAtt-1; l>=0; l--) {
						dist=fabs(new_cov[i][l+j*nbAtt]-cov[i][l+j*nbAtt]);
						if (dist>=EPS_EM) {
							go_on=true;
							break;
						}
					}
					if (go_on) break;
				}
				if (go_on) break;
			}
			// si cet ecart est assez petit, on sort
			if (!go_on) break;
	
			// sinon: probasClassModes=new_probasClassModes, means:=new_means, cov:=new_cov et on recommence
			for (int i=nbModes-1; i>=0; i--) {
				probasClassModes[i].proba=new_probasClassModes[i];
				for (int j=nbAtt-1; j>=0; j--) {
					means[i][j]=new_means[i][j];
					for (int l=nbAtt-1; l>=0; l--) cov[i][l+j*nbAtt]=new_cov[i][l+j*nbAtt];
				}
			}
	
			// on recalcule les nouvelles valeurs des determinants et matrices de covariance inverses:
			initClassify(nbModes);
		}
	}

	// labelisations des modes:
	double min_dist;
	int bestIndex,maxRep,**repartitions=(int**)malloc(nbModes*sizeof(int*));
	for (int i=nbModes-1; i>=0; i--) repartitions[i]=(int*)calloc(nbClasses,sizeof(int));
	for (int i=nbLines_A-1; i>=0; i--) {
		// on calcule les distances aux centroides (==means)
		min_dist=MAX_DOUBLE;
		bestIndex=-1;
		for (int j=nbModes-1; j>=0; j--) {
			dist=0.0;
			count=0;
			for (int l=nbAtt-1; l>=0; l--) {
				switch (typesAttributs[l]) {
					case 's':
						if (!attSymbolic_A[tabIndex[l]][i].miss) {
							dist+=(1.0-distribSymb[j][tabIndex[l]][indexOfSymbol_A[tabIndex[l]][i]]);
							count++;
						}
						break;
					case 'c':
						if (!attContinue_A[tabIndex[l]][i].miss) {
							dist+=fabs(new_attContinue_A[tabIndex[l]][i]-means[j][l]);
							count++;
						}
						break;
					case 'd':
						if (!attDiscrete_A[tabIndex[l]][i].miss) {
							dist+=fabs(new_attDiscrete_A[tabIndex[l]][i]-means[j][l]);
							count++;
						}
						break;
					default: break; // cas de 'i'
				}
			}
			if (count>0) dist/=count;
			if (dist<min_dist) {
				min_dist=dist;
				bestIndex=j;
			}
		}
		repartitions[bestIndex][classIndex_A[i]]++;
	}
	for (int i=nbModes-1; i>=0; i--) {
		maxRep=0;
		bestIndex=-1;
		for (int j=nbClasses-1; j>=0; j--) {
			if (repartitions[i][j]>maxRep) {
				maxRep=repartitions[i][j];
				bestIndex=j;
			}
		}
		if (maxRep==0) probasClassModes[i].label=UNKNOWN_CLASS;
		else probasClassModes[i].label=initData[bestIndex].str;
	}

	// liberations memoire
	free(new_probasClassModes);
	for (int i=nbModes-1; i>=0; i--) {
		free(repartitions[i]);
		free(new_means[i]);
		free(new_cov[i]);
	}
	free(repartitions);
	free(new_means);
	free(new_cov);
}

// classifie l'exemple de la ligne numLine, supposant initialize() appele
string& ClassifGaussMixt::classify(int numLine,bool testData) {
	double proba,bestProba=-1.0;
	int bestIndex=-1;
	for (int i=nbModes-1; i>=0; i--) {
		// on calcule les probabilites de chacun des modes, sachant la donnee
		proba=probasClassModes[i].proba;
		if (determinants[i]>0.0) proba*=(1.0/sqrt(determinants[i]))*exp(-0.5*normeCarree(numLine,i,testData));
		if (proba>bestProba) {
			bestProba=proba;
			bestIndex=i;
		}
	}
	return probasClassModes[bestIndex].label;
}

// affichage des parametres dans un fichier
// remarque: les resultats de calculs avec valeurs inutiles (identifiants) sont simplement a zero
void ClassifGaussMixt::printToFile() {
	ofstream outFile("results/paramGaussMixt.txt");
	bool meansPrinted;
	// on normalise les probabilites:
	double sum=0.0;
	for (int i=nbModes-1; i>=0; i--) sum+=probasClassModes[i].proba;
	if (sum>0.0) for (int i=nbModes-1; i>=0; i--) probasClassModes[i].proba/=sum;
	outFile << "/// moyennes et matrices de covariance des modes: ///" << endl << endl;
	for (int i=0; i<nbModes; i++) {
		meansPrinted=false;
		outFile << "|||| MODE " << (i+1) << " de probabilite " << probasClassModes[i].proba << ", classe " 
		<< probasClassModes[i].label << " ||||" << endl << endl;
		for (int j=0; j<nbAtt; j++) {
			if (!meansPrinted) {
				outFile << "MEAN=" << endl << "   ";
				meansPrinted=true;
			}
			outFile << fixed << means[i][j] << " ";
		}
		if (meansPrinted) {
			outFile << endl;
			outFile << "COV_MATRIX=" << endl;
			for (int j=0; j<nbAtt; j++) {
				outFile << "  ";
				for (int k=0; k<nbAtt; k++) {
					if (cov[i][k+j*nbAtt]>=0.0) outFile << " ";
					outFile << fixed << cov[i][k+j*nbAtt] << " ";
				}
				outFile << endl;
			}
		}
		if (!attSymbolic_A.empty()) {
			if (meansPrinted) outFile << endl;
			outFile << "les probabilites des symboles respectifs suivants: " << endl;
			for (int j=nbSymbols-1; j>=0; j--) outFile << symbols[j] << "  ";
			outFile << endl << "sont, par attribut:" << endl;
		}
		for (int j=0; j<(int)attSymbolic_A.size(); j++) {
			outFile << (j+1) << ":" << endl;
			for (int k=nbSymbols-1; k>=0; k--) outFile << distribSymb[i][j][k] << " ";
			outFile << endl;
		}
		outFile << endl;
		if (!attSymbolic_A.empty()) outFile << endl;
	}
}
