#include "ClassifTree.h"

// tableau des puissances de 2 necessaires a l'acceleration des calculs
unsigned long long* puiss2;

// la fonction de creation de l'arbre de classification
Arbre buildTree(char splitCriteria) {
	// initialisations:
	puiss2=new unsigned long long[nbClasses];
	for (int i=nbClasses-1; i>=0; i--) puiss2[i]=1;
	for (int i=1; i<nbClasses; i++) puiss2[i]=2*puiss2[i-1];
	Arbre abr=Arbre(initData);

	int curID=0;
	abr.racine->calculateSplitValue(splitCriteria);
	Noeud *node,**newFils;
	vector<vector<string> > subsets;
	double cutFloat,valSplitSymb,valSplitCont,valSplitDisc;
	int cutInt,indexAttString,indexAttFloat,indexAttInt,index,nbFilsNV;
	char maximum;

	while (abr.getOpenNode(&node)) { // on choisit arbitrairement un noeud de l'arbre non resolu
									// le critere d'arret pourra [devra..] etre assoupli plus tard
		node->id=curID++;
		valSplitSymb=findBestSplitSymbolic(node,subsets,indexAttString,splitCriteria);
		valSplitCont=findBestSplitContinue(node,cutFloat,indexAttFloat,splitCriteria);
		valSplitDisc=findBestSplitDiscrete(node,cutInt,indexAttInt,splitCriteria);
		// dans le cas ou on n'a pas pu ameliorer le split indice, on continue sans engendrer de fils
		if (valSplitSymb<0.0 && valSplitCont<0.0 && valSplitDisc<0.0) {
			node->type='f';
			node->setLabel();
			continue;
		}
		maximum=maxm(valSplitSymb,valSplitCont,valSplitDisc);

		// on separe les donnees:
		if (maximum=='s') {
			node->type='s'; // jusqu'ici le type n'etait pas connu
			node->indexAtt=indexAttString;
			node->subsets=subsets;
			newFils=new Noeud*[subsets.size()];
			for (int i=subsets.size()-1; i>=0; i--) newFils[i]=new Noeud();
			
			// parcours des exemples du noeud pere pour les distribuer dans les noeuds fils;
			// les missing_value sont ajoutes aux fils avec une distribution de probabilite
			// proportionnelle aux cardinaux des differentes classes presentes (sauf si trop peu significatifs)
			for (int i=node->distribExemples.size()-1; i>=0; i--) {
				for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
					if (!attSymbolic_A[indexAttString][node->distribExemples[i].occurrences[j]].miss) {
						newFils[getIndexIn(subsets,attSymbolic_A[indexAttString][node->distribExemples[i].occurrences[j]].str)]->
							addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
					}
					else {
						// valeur manquante
						index=findIndexSymbolic(node,subsets,indexAttString,classes_A[node->distribExemples[i].occurrences[j]]);
						if (index>=0)
							newFils[index]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
					}
				}
			}
			// ajout des fils au noeud courant, eventuellement calcul des split indices de base + link sur pere
			node->addSuccessors(newFils,subsets.size());
		}
		else {
			node->type=maximum; // jusqu'ici le type n'etait pas connu
			if (maximum=='c') {
				node->indexAtt=indexAttFloat;
				node->cutFloat=cutFloat;
			}
			else {
				node->indexAtt=indexAttInt;
				node->cutInt=cutInt;
			}
			newFils=new Noeud*[2];
			newFils[0]=new Noeud();
			newFils[1]=new Noeud();
			// parcours des exemples du noeud pere pour les distribuer dans les noeuds fils;
			// les missing_value sont ajoutes aux fils avec une distribution de probabilite
			// proportionnelle aux cardinaux des differentes classes presentes (sauf si trop peu significatif)
			for (int i=node->distribExemples.size()-1; i>=0; i--) {
				for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
					if (maximum=='c') {
						if (!attContinue_A[indexAttFloat][node->distribExemples[i].occurrences[j]].miss) {
							if (attContinue_A[indexAttFloat][node->distribExemples[i].occurrences[j]].val<=cutFloat)
								newFils[0]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
							else 
								newFils[1]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
						}
						else {
							// valeur manquante
							index=findIndexContinue(node,cutFloat,indexAttFloat,classes_A[node->distribExemples[i].occurrences[j]]);
							if (index>=0)
								newFils[index]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
						}
					}
					else {
						if (!attDiscrete_A[indexAttInt][node->distribExemples[i].occurrences[j]].miss) {
							if (attDiscrete_A[indexAttInt][node->distribExemples[i].occurrences[j]].val<=cutInt)
								newFils[0]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
							else 
								newFils[1]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
						}
						else {
							// valeur manquante
							index=findIndexDiscrete(node,cutInt,indexAttInt,classes_A[node->distribExemples[i].occurrences[j]]);
							if (index>=0)
								newFils[index]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
						}
					}
				}
			}
			// ajout des fils au noeud courant, eventuellement calcul des split indices de base + link sur pere
			node->addSuccessors(newFils,2);
		}
		// apres split, on met a jour les noeuds ouverts (fils du noeud juste traite)
		// s'il n'y a qu'un fils "non vide", on s'arrete
		nbFilsNV=0;
		index=0;
		for (int i=node->nbFils-1; i>=0; i--) {
			newFils[i]->setFather(node); // pointeur sur pere
			newFils[i]->calculateSplitValue(splitCriteria);
			if (newFils[i]->splitValue>0.0) {
				abr.openNodes.push_back(newFils[i]);
				nbFilsNV++;
				index=i;
			}
			else {
				newFils[i]->id=curID++;
				if (!(newFils[i]->distribExemples.empty())) nbFilsNV++;
				newFils[i]->type='f';
				newFils[i]->setLabel();
			}
		}
		if (nbFilsNV<=1) {
			abr.openNodes.pop_back();
			newFils[index]->type='f';
			newFils[index]->setLabel();
		}
	}
	delete [] puiss2;
	
	return abr;
}

// fonction auxiliaire: trouve les valeurs de split de part et d'autre d'un point de coupure
void computeSplit(vector<FloatAndClass>& valeursAtt,int& nbTotalExemples,int countMissingValues,double f,double& splitGauche,double& splitDroite,double& branchFact,char splitCriteria) {
	double* proportions=(double*)calloc(nbClasses,sizeof(double));
	double proportion,maxProp;
	// calcul du split indice a gauche: on cherche d'abord les proportions
	nbTotalExemples=0;
	int k;
	for (k=0; k<(int)valeursAtt.size(); k++) {
		if (valeursAtt[k].val>f) break;
		nbTotalExemples++;
		proportions[valeursAtt[k].classIndex]+=1.0;
	}
	if (nbTotalExemples==0) splitGauche=0.0; // attention aux divisions par 0!
	else {
		switch (splitCriteria) {
			case 'g':
				splitGauche=1.0;
				for (int i=nbClasses-1; i>=0; i--) splitGauche-=square(proportions[i]/nbTotalExemples);
				break;
			case 'e':
				splitGauche=0.0;
				for (int i=nbClasses-1; i>=0; i--) {
					proportion=proportions[i]/nbTotalExemples;
					if (proportion>0) splitGauche-=proportion*log(proportion);
				}
				break;
			case 'm':
				splitGauche=1.0;
				maxProp=0.0;
				for (int i=nbClasses-1; i>=0; i--) {
					proportion=proportions[i]/nbTotalExemples;
					if (proportion>maxProp) maxProp=proportion;
				}
				splitGauche-=maxProp;
				break;
			default: 
				splitGauche=0.0; // jamais atteint mais evite les warnings
				break;
		}
	}

	// calcul du split indice a droite: on cherche d'abord les proportions
	for (int i=nbClasses-1; i>=0; i--) proportions[i]=0.0; // remise a zero des proportions
	nbTotalExemples=0;
	for (; k<(int)valeursAtt.size(); k++) {
		nbTotalExemples++;
		proportions[valeursAtt[k].classIndex]+=1.0;
	}
	if (nbTotalExemples==0) splitDroite=0.0; // attention aux divisions par 0!
	else {
		switch (splitCriteria) {
			case 'g':
				splitDroite=1.0;
				for (int i=nbClasses-1; i>=0; i--) splitDroite-=square(proportions[i]/nbTotalExemples);
				break;
			case 'e':
				splitDroite=0.0;
				for (int i=nbClasses-1; i>=0; i--) {
					proportion=proportions[i]/nbTotalExemples;
					if (proportion>0) splitDroite-=proportion*log(proportion);
				}
				break;
			case 'm':
				splitDroite=1.0;
				maxProp=0.0;
				for (int i=nbClasses-1; i>=0; i--) {
					proportion=proportions[i]/nbTotalExemples;
					if (proportion>maxProp) maxProp=proportion;
				}
				splitDroite-=maxProp;
				break;
			default: 
				splitDroite=0.0; // jamais atteint mais evite les warnings
				break;
		}
	}
	free(proportions);

	// facteur de normalisation (ameliore les perfs de l'entropie par exemple)
	branchFact=0.0;
	if (valeursAtt.size()-nbTotalExemples>0) {
		proportion=(double)(valeursAtt.size()-nbTotalExemples)/(valeursAtt.size()+countMissingValues);
		branchFact-=proportion*log(proportion);
	}
	if (nbTotalExemples>0) {
		proportion=(double)nbTotalExemples/(valeursAtt.size()+countMissingValues);
		branchFact-=proportion*log(proportion);
	}
}

// des fonctions auxiliaires qui retourne la meilleure valeur de split
// retourne -1 si type d'attribut non present
// on utilise la methode "reduction d'entropie/Gini/misc_error" divisee par split_ratio
double findBestSplitSymbolic(Noeud* node,vector<vector<string> >& subsets,int& indexAtt,char splitCriteria) {
	if (attSymbolic_A.empty()) return -1.0;
	vector<vector<string> > curSubsets;
	vector<StringAndClassesID> chainesClasses; // chaines (symboles) rencontrees au temps + classes correspondantes
	double bestSplitVal=-1.0,proportion,maxProp;
	double* proportions=new double[nbClasses];
	int countMissingValues; // valeurs manquantes
	// methode: on split sur toutes les valeurs discretes puis on fait des regroupements eventuels
	for (int i=attSymbolic_A.size()-1; i>=0; i--) {
		// tentative pour l'attribut symbolique numero i

		// recherche de tous les symboles presents pour l'attribut i: un vecteur par symbole pour commencer
		curSubsets.clear();
		chainesClasses.clear();
		countMissingValues=0;
		for (int j=node->distribExemples.size()-1; j>=0; j--) {
			for (int k=node->distribExemples[j].occurrences.size()-1; k>=0; k--) {
				if (!attSymbolic_A[i][node->distribExemples[j].occurrences[k]].miss) {
					int l;
					for (l=chainesClasses.size()-1; l>=0; l--) 
						if (attSymbolic_A[i][node->distribExemples[j].occurrences[k]].str==chainesClasses[l].str) break;
					if (l>=0) continue;
					vector<string> vs;
					vs.push_back(attSymbolic_A[i][node->distribExemples[j].occurrences[k]].str);
					curSubsets.push_back(vs);
					unsigned long long classesID=0;
					bool* pbool=(bool*)calloc(nbClasses,sizeof(bool));
					chainesClasses.push_back(StringAndClassesID(attSymbolic_A[i][node->distribExemples[j].occurrences[k]].str,classesID,pbool));
				}
				else countMissingValues++;
			}
		}

		// on regarde si des regroupements sont possibles:
		// on associe un "vecteur de bits" a chaque chaine trouvee, representant les classes qui ont cette valeur d'attribut.
		// dans le cas ou deux chaines ont le meme vecteur de bits, il faut les fusionner;
		// pour l'instant c'est le seul cas de fusion que l'on envisage (sinon complexite exponentielle)
		for (int j=node->distribExemples.size()-1; j>=0; j--) {
			for (int k=node->distribExemples[j].occurrences.size()-1; k>=0; k--) {
				// si la valeur de l'attribut est manquante, on continue:
				if (attSymbolic_A[i][node->distribExemples[j].occurrences[k]].miss) continue;
				int l;
				for (l=chainesClasses.size()-1; l>=0; l--) 
					if (chainesClasses[l].str==attSymbolic_A[i][node->distribExemples[j].occurrences[k]].str) break;
				if (!chainesClasses[l].pbool[node->distribExemples[j].index]) {
					chainesClasses[l].classesID+=puiss2[node->distribExemples[j].index];
					chainesClasses[l].pbool[node->distribExemples[j].index]=true;
				}
			}
		}

		// plus qu'a apparier les chaines aux classesID similaires, SAUF si un seul classesID (==on ne ferait rien en regroupant tout!)
		unsigned long long valTest=chainesClasses.back().classesID; // necessairement valTest>0 dans au moins une classe
		for (int j=chainesClasses.size()-2; j>=0; j--) {
			if (chainesClasses[j].classesID!=valTest) {
				valTest=0;
				break;
			}
		}
		if (valTest<=0) {
			bool* chaineDejaAppariee=(bool*)calloc(chainesClasses.size(),sizeof(bool)); // memorisation des chaines traitees
			for (int j=0; j<(int)chainesClasses.size(); j++) {
				for (int k=j+1; k<(int)chainesClasses.size(); k++) {
					if (!chaineDejaAppariee[k]) {
						if (chainesClasses[k].classesID==chainesClasses[j].classesID) {
							curSubsets[j].push_back(chainesClasses[k].str);
							curSubsets[k].pop_back();
							chaineDejaAppariee[k]=true;
						}
					}
				}
			}
			free(chaineDejaAppariee);

			// on supprime les cellules vides en fin du vecteur curSubsets
			for (int j=0; j<(int)curSubsets.size(); j++) {
				if (curSubsets[j].empty()) {
					// on echange avec le premier sous-ensemble non vide que l'on trouve a partir de la (s'il en existe..)
					int k;
					for (k=j+1; k<(int)curSubsets.size(); k++)
						if (!curSubsets[k].empty()) break;
					if (k>=(int)curSubsets.size()) break;
					while (!curSubsets[k].empty()) {
						curSubsets[j].push_back(curSubsets[k].back());
						curSubsets[k].pop_back();
					}
				}
			}
			while (!curSubsets.empty() && curSubsets.back().empty()) curSubsets.pop_back();
		}

		// calcul de l'indice de discrimination avec curSubsets (attention, gain relatif d'information, on maximise!)
		int nbTotalExemples=0;
		int* nbExemples=(int*)calloc(curSubsets.size(),sizeof(int));
		double* splitIndice=new double[curSubsets.size()];
		for (int j=curSubsets.size()-1; j>=0; j--) {
			// pour chaque sous-ensemble il faut calculer le split indice, puis mixer le tout comme d'hab
			for (int k=nbClasses-1; k>=0; k--) proportions[k]=0;
			for (int k=node->distribExemples.size()-1; k>=0; k--) {
				for (int l=node->distribExemples[k].occurrences.size()-1; l>=0; l--) {
					if (isIn(attSymbolic_A[i][node->distribExemples[k].occurrences[l]].str,curSubsets[j])) {
						proportions[node->distribExemples[k].index]+=1.0;
						nbExemples[j]++;
						nbTotalExemples++;
					}
				}
			}
			// calcul de splitIndice[j]
			if (nbExemples[j]==0) {
				splitIndice[j]=0.0;
				continue;
			}
			switch (splitCriteria) {
				case 'g':
					splitIndice[j]=1.0;
					for (int k=nbClasses-1; k>=0; k--) splitIndice[j]-=square(proportions[k]/nbExemples[j]);
					break;
				case 'e':
					splitIndice[j]=0.0;
					for (int k=nbClasses-1; k>=0; k--) {
						proportion=proportions[k]/nbExemples[j];
						if (proportion>0) splitIndice[j]-=proportion*log(proportion);
					}
					break;
				case 'm':
					splitIndice[j]=1.0;
					maxProp=0.0;
					for (int k=nbClasses-1; k>=0; k--) {
						proportion=proportions[k]/nbExemples[j];
						if (proportion>maxProp) maxProp=proportion;
					}
					splitIndice[j]-=maxProp;
					break;
				default: 
					splitIndice[j]=0.0; // jamais atteint mais evite les warnings
					break;
			}
		}
		// calcul de splitVal avec les splitIndice[j] et les nbExemples[j] + nombre total d'exemples
		// encore une fois, attention a ne pas diviser par 0
		double splitVal,tmpVal,branchFact;
		if (splitCriteria!='m') splitVal=node->splitValue;
		else splitVal=1.0; // car trop souvent la valeur recomposee est tres proche de celle du pere
		tmpVal=0.0;
		for (int j=curSubsets.size()-1; j>=0; j--) tmpVal+=nbExemples[j]*splitIndice[j];
		tmpVal/=(nbTotalExemples+countMissingValues);
		
		// on poursuit seulement si amelioration par rapport au split indice du pere
		if (tmpVal<node->splitValue) {
			splitVal-=tmpVal; // ici on est sur que nbTotalExemples>0, car node dans open
			// normalisation par le facteur de branchement:
			branchFact=0.0;
			for (int j=curSubsets.size()-1; j>=0; j--) {
				proportion=(double)nbExemples[j]/(nbTotalExemples+countMissingValues);
				if (proportion>0) branchFact-=proportion*log(proportion);
			}
			if (branchFact>0.0) splitVal/=branchFact;
			else splitVal-=1.0; // on ne veut pas des attributs non discriminants!
			if (splitVal>bestSplitVal) {
				bestSplitVal=splitVal;
				subsets=curSubsets;
				indexAtt=i;
			}
		}

		// liberation memoire
		free(nbExemples);
		delete [] splitIndice;
	}
	
	delete [] proportions;
	return bestSplitVal;
}

double findBestSplitContinue(Noeud* node,double& cutFloat,int& indexAtt,char splitCriteria) {
	if (attContinue_A.empty()) return -1.0;
	// on commence par trier les attributs par valeurs croissantes,
	// puis on teste de l'ordre de sqrt(n) splits differents equirepartis
	// on garde le meilleur (tous attributs continus confondus), tenant compte des missing_values
	double bestSplitVal=-1.0;
	vector<FloatAndClass> valeursAtt;
	for (int i=attContinue_A.size()-1; i>=0; i--) {
		// tentative pour l'attribut continu numero i

		// a) recuperation des valeurs des attributs
		int countMissingValues=0;
		valeursAtt.clear();
		for (int j=node->distribExemples.size()-1; j>=0; j--) {
			for (int k=node->distribExemples[j].occurrences.size()-1; k>=0; k--) {
				if (!attContinue_A[i][node->distribExemples[j].occurrences[k]].miss)
					valeursAtt.push_back(FloatAndClass(attContinue_A[i][node->distribExemples[j].occurrences[k]].val,node->distribExemples[j].index));
				else countMissingValues++;
			}
		}
		// tri des valeurs
		sort(valeursAtt.begin(),valeursAtt.end());

		// b) tests de splits
		double range=valeursAtt.back().val-valeursAtt[0].val; // amplitude de variation de l'attribut
		double step=range/sqrt((double)valeursAtt.size()); // ecart entre deux tests de splits
		for (double f=valeursAtt[0].val+step/2.0; f<valeursAtt.back().val; f+=step) {

			// recherche des valeurs de split a droite et a gauche du point de coupure
			double splitGauche,splitDroite,branchFact;
			int nbTotalExemples;
			computeSplit(valeursAtt,nbTotalExemples,countMissingValues,f,splitGauche,splitDroite,branchFact,splitCriteria);
			
			double splitVal;
			if (splitCriteria!='m') splitVal=node->splitValue;
			else splitVal=1.0; // car trop souvent la valeur recomposee est tres proche de celle du pere
			double tmpVal=(((double)(valeursAtt.size()-nbTotalExemples)/(valeursAtt.size()+countMissingValues))*splitGauche+((double)nbTotalExemples/(valeursAtt.size()+countMissingValues))*splitDroite);

			// on poursuit seulement si amelioration par rapport au split indice du pere
			if (tmpVal<node->splitValue) {
				splitVal-=tmpVal;
				if (branchFact>0.0) splitVal/=branchFact;
				else splitVal-=1.0; // on ne veut pas des attributs non discriminants!
				if (splitVal>bestSplitVal) {
					bestSplitVal=splitVal;
					cutFloat=f;
					indexAtt=i;
				}
			}
		}
	}

	return bestSplitVal;
}
double findBestSplitDiscrete(Noeud* node,int& cutInt,int& indexAtt,char splitCriteria) {
	if (attDiscrete_A.empty()) return -1.0;
	// on commence par trier les attributs par valeurs croissantes,
	// puis on teste de l'ordre de sqrt(n) splits differents equirepartis
	// on garde le meilleur (tous attributs continus confondus), tenant compte des missing_values
	double bestSplitVal=-1.0;
	vector<FloatAndClass> valeursAtt;
	for (int i=attDiscrete_A.size()-1; i>=0; i--) {
		// tentative pour l'attribut discret numero i

		// a) recuperation des valeurs des attributs
		int countMissingValues=0;
		valeursAtt.clear();
		for (int j=node->distribExemples.size()-1; j>=0; j--) {
			for (int k=node->distribExemples[j].occurrences.size()-1; k>=0; k--) {
				if (!attDiscrete_A[i][node->distribExemples[j].occurrences[k]].miss)
					valeursAtt.push_back(FloatAndClass(attDiscrete_A[i][node->distribExemples[j].occurrences[k]].val,node->distribExemples[j].index));
				else countMissingValues++;
			}
		}
		// tri des valeurs
		sort(valeursAtt.begin(),valeursAtt.end());

		// b) tests de splits
		double range=valeursAtt.back().val-valeursAtt[0].val; // amplitude de variation de l'attribut
		double step=maxm(1.0,range/sqrt((double)valeursAtt.size())); // ecart entre deux tests de splits
		for (double f=valeursAtt[0].val+step/2.0; f<valeursAtt.back().val; f+=step) {

			// recherche des valeurs de split a droite et a gauche du point de coupure
			double splitGauche,splitDroite,branchFact;
			int nbTotalExemples;
			computeSplit(valeursAtt,nbTotalExemples,countMissingValues,f,splitGauche,splitDroite,branchFact,splitCriteria);

			double splitVal;
			if (splitCriteria!='m') splitVal=node->splitValue;
			else splitVal=1.0; // car trop souvent la valeur recomposee est tres proche de celle du pere
			double tmpVal=(((double)(valeursAtt.size()-nbTotalExemples)/(valeursAtt.size()+countMissingValues))*splitGauche+((double)nbTotalExemples/(valeursAtt.size()+countMissingValues))*splitDroite);
			
			// on poursuit seulement si amelioration par rapport au split indice du pere
			if (tmpVal<node->splitValue) {
				splitVal-=tmpVal;
				if (branchFact>0.0) splitVal/=branchFact;
				else splitVal-=1.0; // on ne veut pas des attributs non discriminants!
				if (splitVal>bestSplitVal) {
					bestSplitVal=splitVal;
					cutInt=(int)f;
					indexAtt=i;
				}
			}
		}
	}

	return bestSplitVal;
}

// fonctions pour determiner ou va aller une valeur manquante
int findIndexSymbolic(Noeud* node,const vector<vector<string> >& subsets,int indexAtt,const string& str) {
	for (int i=node->distribExemples.size()-1; i>=0; i--) {
		if (node->distribExemples[i].str==str) {
			// on initialise les repartitions des classes
			int* repartitions=(int*)calloc(subsets.size(),sizeof(int));	
			int index,sumRep=0,randRes;
			for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
				if (!attSymbolic_A[indexAtt][node->distribExemples[i].occurrences[j]].miss) {
					index=getIndexIn(subsets,attSymbolic_A[indexAtt][node->distribExemples[i].occurrences[j]].str);
					if (index>=0) {
						repartitions[index]++;
						sumRep++;
					}
				}
			}
			for (int j=subsets.size()-1; j>=0; j--) {
				if (repartitions[j]>=Arbre::NBMIN_REPDISC) {
					randRes=rand()%sumRep+1;
					sumRep=repartitions[0];
					for (int k=0; k<(int)subsets.size(); k++) {
						if (randRes<=sumRep) return k;
						sumRep+=repartitions[k+1];
					}
				}
			}
			free(repartitions);
			break;
		}
	}
	// cas ou on ne trouve pas assez d'elements significatifs
	/*** extension future:
	 *** definir une metrique sur l'espace des attributs,
	 *** et chercher les classes des exemples les plus proches au sens de cette metrique. */
	return -1;
}
int findIndexContinue(Noeud* node,double cutFloat,int indexAtt,const string& str) {
	for (int i=node->distribExemples.size()-1; i>=0; i--) {
		if (node->distribExemples[i].str==str) {
			int aGauche=0,aDroite=0;
			for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
				if (!attContinue_A[indexAtt][node->distribExemples[i].occurrences[j]].miss) {
					if (attContinue_A[indexAtt][node->distribExemples[i].occurrences[j]].val<=cutFloat) aGauche++;
					else aDroite++;
				}
			}
			if (aGauche>=Arbre::NBMIN_REPBIN || aDroite>=Arbre::NBMIN_REPBIN) {
				if (rand()%(aGauche+aDroite)+1<=aGauche) return 0;
				else return 1;
			}
			break;
		}
	}
	// cas ou on ne trouve pas assez d'elements significatifs
	/*** extension future:
	 *** definir une metrique sur l'espace des attributs,
	 *** et chercher les classes des exemples les plus proches au sens de cette metrique. */
	return -1;
}
int findIndexDiscrete(Noeud* node,int cutInt,int indexAtt,const string& str) {
	for (int i=node->distribExemples.size()-1; i>=0; i--) {
		if (node->distribExemples[i].str==str) {
			int aGauche=0,aDroite=0;
			for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
				if (!attDiscrete_A[indexAtt][node->distribExemples[i].occurrences[j]].miss) {
					if (attDiscrete_A[indexAtt][node->distribExemples[i].occurrences[j]].val<=cutInt) aGauche++;
					else aDroite++;
				}
			}
			if (aGauche>=Arbre::NBMIN_REPBIN || aDroite>=Arbre::NBMIN_REPBIN) {
				if (rand()%(aGauche+aDroite)+1<=aGauche) return 0;
				else return 1;
			}
			break;
		}
	}
	// cas ou on ne trouve pas assez d'elements significatifs
	/*** extension future:
	 *** definir une metrique sur l'espace des attributs,
	 *** et chercher les classes des exemples les plus proches au sens de cette metrique. */
	return -1;
}
