#ifndef _LoadData
#define _LoadData

#include <string>
#include <vector>
#include <string>
#include <limits>
#include <cmath>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include "Functions.h"

using namespace std;

// une structure utile avec une chaine et un index de classe + compteur du nombre d'occurences de la classe + ces occurrences
	
struct StringIndexCounter {
	StringIndexCounter () {
		counter=0;
	}
	StringIndexCounter (const string& s,int i,int c) {
		str=s;
		index=i;
		counter=c;
	}
	string str;
	int index,counter;
	vector<int> occurrences;
};

// types unions pour prendre en compte les missing values

struct StringOuMiss {
	StringOuMiss () {
		miss=false;
	}
	string str;
	bool miss;
};
struct FloatOuMiss {
	FloatOuMiss () {
		miss=false;
		isInit=false;
	}
	double val;
	bool miss;
	bool isInit; // utile pour ajuster le nombre de lignes
};
struct IntOuMiss {
	IntOuMiss () {
		miss=false;
		isInit=false;
	}
	int val;
	bool miss;
	bool isInit; // utile uniquement pour les tests
};

// structure contenant un label de classe + probabilite

struct ClassAndProba {
	ClassAndProba () {}
	ClassAndProba (string& str,double prob) {
		label=str;
		proba=prob;
	}
	string label;
	double proba;
};

// structure contenant un symbole + probabilite (identique a la precedente, separee pour plus de clarte
// car les usages de ces deux structures sont tres differents; celle-ci est utile pour les calculs de
// distances entre attributs symboliques numerises)

struct SymbolAndProba {
	SymbolAndProba () {}
	SymbolAndProba (string& sym,double prob) {
		symb=sym;
		proba=prob;
	}
	string symb;
	double proba;
};

// index + proba: utile dans ClassifGaussMixt pour les tris

struct IndexAndProba {
	IndexAndProba () {}
	IndexAndProba (int ind,double prob) {
		index=ind;
		proba=prob;
	}
	bool operator<(const IndexAndProba& indPro) const {
		return proba<indPro.proba;
	}
	int index;
	double proba;
};

class LoadData {

public:
	
	LoadData () {}
	
	~LoadData () {}
	
	// chargement des donnees (d'entrainement ou de test)
	static void loadFromFile(char* name,const string& typesAttributs,string dataType);

	// recuperation des distributions de probabilite des donnees symboliques
	static void getDistribSymbolic();
	
	// normalisation des donnees continues/discretes (utiles pour calculer des distances)
	static void normalizeData();

	// les attributs ranges par types:
	static vector<StringOuMiss*> attSymbolic_A,attSymbolic_T;
	static vector<FloatOuMiss*> attContinue_A,attContinue_T;
	static vector<IntOuMiss*> attDiscrete_A,attDiscrete_T;
	static int nbClasses; // nombre total de classes (obtenu sur les donnees d'entrainement)
	static string *classes_A,*classes_T; // les chaines decrivant les classes, par ligne
	static int nbLines_A,nbLines_T; // nombre d'exemples
	static int* tabIndex; // indique les index des types d'attributs dans leurs tableaux respectifs
	
	// index des classes dans initData en fonction de la ligne;
	// utile dans k-nn par exemple
	static int *classIndex_A,*classIndex_T;
	// pour initialiser le classifieur: donnees rangees dans un StringIndexCounter:
	// seulement sur donnees d'entrainement
	static vector<StringIndexCounter> initData;
	static vector<string> symbols; // les chaines des symboles presents
	static int **indexOfSymbol_A,**indexOfSymbol_T; // les index (==identifiants) des symboles
	static int nbSymbols; // le nombre total de symboles

	// atributs normalises
	static double **new_attContinue_A,**new_attContinue_T;
	static double **new_attDiscrete_A,**new_attDiscrete_T;
	
	// nombre d'attributs non identifiants:
	static int nbRealAtt;
	
	// constante max_double:
	static constexpr double MAX_DOUBLE=1e100;
	
	// chaine "classe inconnue"
	static string UNKNOWN_CLASS;
	
};

#define loadFromFile LoadData::loadFromFile
#define getDistribSymbolic LoadData::getDistribSymbolic
#define normalizeData LoadData::normalizeData

#define attSymbolic_A LoadData::attSymbolic_A
#define attSymbolic_T LoadData::attSymbolic_T
#define attContinue_A LoadData::attContinue_A
#define attContinue_T LoadData::attContinue_T
#define attDiscrete_A LoadData::attDiscrete_A
#define attDiscrete_T LoadData::attDiscrete_T
#define nbClasses LoadData::nbClasses
#define classes_A LoadData::classes_A
#define classes_T LoadData::classes_T
#define nbLines_A LoadData::nbLines_A
#define nbLines_T LoadData::nbLines_T
#define tabIndex LoadData::tabIndex

#define classIndex_A LoadData::classIndex_A
#define classIndex_T LoadData::classIndex_T
#define initData LoadData::initData
#define symbols LoadData::symbols
#define indexOfSymbol_A LoadData::indexOfSymbol_A
#define indexOfSymbol_T LoadData::indexOfSymbol_T
#define nbSymbols LoadData::nbSymbols
#define new_attContinue_A LoadData::new_attContinue_A
#define new_attContinue_T LoadData::new_attContinue_T
#define new_attDiscrete_A LoadData::new_attDiscrete_A
#define new_attDiscrete_T LoadData::new_attDiscrete_T

#define nbRealAtt LoadData::nbRealAtt
#define MAX_DOUBLE LoadData::MAX_DOUBLE
#define UNKNOWN_CLASS LoadData::UNKNOWN_CLASS

#endif
