#include "MultiDimClassif.h"

// fonctions auxiliaires pour realiser la classification:
double* MultiDimClassif::pivotGauss(double* matrice) {
	// la matrice produit des matrices elementaires intervenant dans le calcul:
	double* prodElemMatrix=(double*)calloc(nbAtt*nbAtt,sizeof(double));
	for (int i=nbAtt-1; i>=0; i--) prodElemMatrix[i+i*nbAtt]=1.0;

	int numCol=0; // index de colonne courant
	int numLine=-1; // numero de ligne pour realiser les echanges/divisions

	double minPos,aux,coeff;
	while (numCol<nbAtt) {
		minPos=0.0; // le plus petit element (en valeur absolue) non nul de sa colonne;
					// choix pour stabilite numerique
		for (int i=numCol; i<nbAtt; i++) {
			if (matrice[numCol+i*nbAtt]!=0.0 && (minPos==0.0 || fabs(matrice[numCol+i*nbAtt])<fabs(minPos))) {
				minPos=matrice[numCol+i*nbAtt];
				numLine=i;
			}
		}
		if (minPos==0.0) { // ctte colonne est deja OK (cas d'un attribut identifiant)
			numCol++;
			continue;
		}

		// il faut echanger puis soustraire les lignes avec les bons coefficients:
		// a) echange des lignes numCol et numLine si celles-ci different:
		if (numCol!=numLine) {
			for (int i=nbAtt-1; i>=0; i--) {
				aux=matrice[i+numCol*nbAtt];
				matrice[i+numCol*nbAtt]=matrice[i+numLine*nbAtt];
				matrice[i+numLine*nbAtt]=aux;
				// on effectue la meme operation sur la matrice produit de matrices elementaires:
				aux=prodElemMatrix[i+numCol*nbAtt];
				prodElemMatrix[i+numCol*nbAtt]=prodElemMatrix[i+numLine*nbAtt];
				prodElemMatrix[i+numLine*nbAtt]=aux;
			}
		}
		// b) soustraction lignes numCol+a - coeff*ligne numCol (pivot de Gauss)
		for (int i=numCol+1; i<nbAtt; i++) {
			if (matrice[numCol+i*nbAtt]==0.0) continue;
			coeff=matrice[numCol+i*nbAtt]/matrice[numCol+numCol*nbAtt];
			for (int j=nbAtt-1; j>=0; j--) {
				matrice[j+i*nbAtt]-=coeff*matrice[j+numCol*nbAtt];
				// on effectue la meme operation sur la matrice produit de matrices elementaires:
				prodElemMatrix[j+i*nbAtt]-=coeff*prodElemMatrix[j+numCol*nbAtt];
			}
		}
		numCol++; // on passe a la colonne suivante et on recommence
	}

	return prodElemMatrix;
}

// PRECONDITION: matrice triangulaire
double MultiDimClassif::determinant(double* matrice) {
	// le resultat est simplement le produit des elements diagonaux non nuls:
	// (certains peuvent etre nuls s'il y a des attributs identifiants, que l'on ignore)
	double res=1.0;
	for (int i=nbAtt-1; i>=0; i--) {
		if (matrice[i+i*nbAtt]!=0.0) res*=matrice[i+i*nbAtt];
	}
	
	return res;
}

double* MultiDimClassif::multMatrix(double* m1,double* m2) {
	double* prod=(double*)calloc(nbAtt*nbAtt,sizeof(double));
	for (int i=nbAtt-1; i>=0; i--) {
		for (int j=nbAtt-1; j>=0; j--) {
			// calcul de prod[i,j]
			for (int k=nbAtt-1; k>=0; k--)
				prod[j+i*nbAtt]+=m1[k+i*nbAtt]*m2[j+k*nbAtt];
		}
	}

	free(m1);
	return prod;
}

double MultiDimClassif::normeCarree(int numLine,int indexClass,bool testData) {
	// on calcule les produits menant au resultat (== t(x-mu)(Sigma-1)(x-mu)):
	// x-mu_i=
	double* xMoinsMui=(double*)malloc(nbAtt*sizeof(double));
	for (int i=nbAtt-1; i>=0; i--) {
		switch (typesAttributs[i]) {
			case 's':
				if ((testData && attSymbolic_T[tabIndex[i]][numLine].miss)
					|| (!testData && attSymbolic_A[tabIndex[i]][numLine].miss)) {
					xMoinsMui[i]=0.0;
					break;
				}
				if (testData) xMoinsMui[i]=1.0-distribSymb[indexClass][tabIndex[i]][indexOfSymbol_T[tabIndex[i]][numLine]];
				else xMoinsMui[i]=1.0-distribSymb[indexClass][tabIndex[i]][indexOfSymbol_A[tabIndex[i]][numLine]];
				break;
			case 'c':
				if ((testData && attContinue_T[tabIndex[i]][numLine].miss)
					|| (!testData && attContinue_A[tabIndex[i]][numLine].miss)) {
					xMoinsMui[i]=0.0;
					break;
				}
				if (testData) xMoinsMui[i]=new_attContinue_T[tabIndex[i]][numLine]-means[indexClass][i];
				else xMoinsMui[i]=new_attContinue_A[tabIndex[i]][numLine]-means[indexClass][i];
				break;
			case 'd':
				if ((testData && attDiscrete_T[tabIndex[i]][numLine].miss)
					|| (!testData && attDiscrete_A[tabIndex[i]][numLine].miss)) {
					xMoinsMui[i]=0.0;
					break;
				}
				if (testData) xMoinsMui[i]=new_attDiscrete_T[tabIndex[i]][numLine]-means[indexClass][i];
				else xMoinsMui[i]=new_attDiscrete_A[tabIndex[i]][numLine]-means[indexClass][i];
				break;
			default: // cas de l'identifiant 'i'
				xMoinsMui[i]=0.0;
				break;
		}
	}

	double* resInterm=(double*)calloc(nbAtt,sizeof(double));
	for (int i=nbAtt-1; i>=0; i--) {
		for (int j=nbAtt-1; j>=0; j--)
			resInterm[i]+=inv_cov[indexClass][j+i*nbAtt]*xMoinsMui[j];
	}

	double res=0.0;
	for (int i=nbAtt-1; i>=0; i--) res+=xMoinsMui[i]*resInterm[i];

	free(xMoinsMui);
	free(resInterm);
	return res;
}

// initialise la classification en calculant les determinants et inverses necessaires
void MultiDimClassif::initClassify(int size) {
	double *prodElemMatrix,*trig_cov=(double*)malloc(nbAtt*nbAtt*sizeof(double));
	for (int i=size-1; i>=0; i--) {
		// on calcule le determinant de la matrice de covariance cov[i] par pivot de Gauss,
		// et l'inverse (ignorant lignes et colonnes nulles car correspondent a identifiants)
		// de cette meme matrice par la methode du systeme.
		for (int j=nbAtt*nbAtt-1; j>=0; j--) trig_cov[j]=cov[i][j];
		prodElemMatrix=pivotGauss(trig_cov);
		determinants[i]=fabs(determinant(trig_cov));
		
		// on resout le systeme triangulaire forme par la matrice trig_cov:
		for (int j=nbAtt-1; j>=0; j--) {
			for (int k=nbAtt-1; k>=0; k--) {
				if (k==j && trig_cov[j+j*nbAtt]!=0.0) inv_cov[i][j+j*nbAtt]=1.0;
				else inv_cov[i][k+j*nbAtt]=0.0;
			}
		}
		for (int line=nbAtt-1; line>=0; line--) {
			if (trig_cov[line+line*nbAtt]==0.0) continue;
			for (int j=nbAtt-1; j>line; j--) {
				for (int k=nbAtt-1; k>=j; k--)
					inv_cov[i][k+line*nbAtt]-=trig_cov[j+line*nbAtt]*inv_cov[i][k+j*nbAtt];
			}
			for (int j=nbAtt-1; j>=line; j--)
				inv_cov[i][j+line*nbAtt]/=trig_cov[line+line*nbAtt];
		}

		// puis on multiplie a droite par prodElemMatrix pour obtenir le resultat correct:
		inv_cov[i]=multMatrix(inv_cov[i],prodElemMatrix);
		free(prodElemMatrix);
	}
	free(trig_cov);
}
