#ifndef _Noeud
#define _Noeud

#include <string>
#include <vector>
#include <cmath>
#include <iostream>
#include "LoadData.h"
#include "Functions.h"

using namespace std;

// la classe d'un noeud/feuille: discrimination sur ensemble discret de chaines string, ou valeur discrete/continue

class Noeud {
	
public:
	
	Noeud () {
		nbFils=0;
		fils=0;
		pere=0;
		dejaTraite=false;
	}
	
	// constructeur avec les differentes classes presentes a ce niveau
	Noeud (vector<StringIndexCounter>& dEx) {
		distribExemples=dEx;
		nbFils=0;
		fils=0;
		pere=0;
		dejaTraite=false;
	}
	
	~Noeud () {
		// on supprime juste les fils: ainsi recursivement a partir de la racine tout disparait
		for (int i=nbFils-1; i>=0; i--) delete fils[i];
		delete [] fils;
	}

	// ajout des fils
	void addSuccessors(Noeud** nodes,int nbF) {
		fils=nodes;
		nbFils=nbF;
	}
	// ajout du pere (utile seulement au pruning)
	void setFather(Noeud* node) {
		pere=node;
	}
	
	// ajout d'une entree dans les exemples restant au niveau de ce noeud
	void addInStringIndexCounter(const string& str,int index,int numLigne);
	
	// calcul de l'indice de discrimination
	void calculateSplitValue(char splitCriteria);
	
	// labelisation par la chaine decrivant la classe dans le cas d'une feuille
	// en cas de conflit de classes, on prend celle qui est majoritaire
	void setLabel() {
		int bestNbInClass=0;
		for (int i=distribExemples.size()-1; i>=0; i--) {
			if (distribExemples[i].counter>bestNbInClass) {
				label=distribExemples[i].str;
				bestNbInClass=distribExemples[i].counter;
			}
		}
		if (bestNbInClass==0)
			label=UNKNOWN_CLASS; // label special dans le cas ou il n'y a aucun exemple
								// ce cas survient lorsque des valeurs symboliques ne sont pas presentes;
								// ou alors dans le cas rare ou tous les attributs entiers/continus 
								// sont egaux sur un noeud
	}
	
	char type; // type du noeud/feuille ('f' pour feuille,'d','c' ou 'o' sinon suivant le type de split)
	vector<StringIndexCounter> distribExemples; // nombres d'elements des differentes classes restant a ce noeud + index..
	Noeud** fils; // noeuds fils eventuels
	int nbFils; // nombre de fils
	Noeud* pere; // pere, utile pour remonter dans l'arbre dans la derniere phase
	double splitValue; // indice de discrimination du noeud
	string label; // label, si feuille
	// sous-ensembles de discrimination suivant le type:
	vector<vector<string> > subsets;
	double cutFloat;
	int cutInt;
	// indice de l'attribut choisi pour split
	int indexAtt;
	
	// utile uniquement pour pruning: booleen indiquant si le pere d'une feuille a ete traite;
	// et identificateur unique d'un noeud
	bool dejaTraite;
	int id;
	
	// utile uniquement pour l'affichage de l'arbre: profondeur du noeud, et sa position sur l'ecran
	int prof,x,y;
	
};

#endif

