#include "Prune.h"

// lagage de l'arbre
void pruning(Arbre& abr,bool fileTest) {
	srand(time(NULL));	
	// on descend dans l'arbre jusqu'aux feuilles, puis on essaye de remonter
	// en fusionnant les feuilles descendant d'un meme noeud;
	// si cela ameliore l'erreur globale de generalisation, alors on fusionne effectivement.
	
	// 2 cas de figure:
	// - on n'a pas de jeu de test; alors on applique simplement la methode 
	//	 "erreur de generalisation pessimiste", qui semble assez adequate
	// - on dispose d'un jeu de test; alors cherche a minimiser la somme des 
	//	 taux d'erreurs realisees.

	vector<Noeud*> nodesToExplore;
	vector<Noeud*> feuilles; // un seul exemplaire par noeud pour faciliter la fusion
	int nbLeafs=0; // vrai nombre de feuilles
	Noeud* curNode; // le noeud en cours d'exploration
	nodesToExplore.push_back(abr.racine);
	// parcours de l'arbre pour trouver les feuilles: on les memorise (une seule feuille maxi par noeud),
	// et on calcule l'erreur de generalisation dans le cas du jeu de test absent
	while (!nodesToExplore.empty()) {
		curNode=nodesToExplore.back();
		nodesToExplore.pop_back();
		if (curNode->type=='f') {
			nbLeafs++;
			if (!curNode->pere->dejaTraite) {
				feuilles.push_back(curNode);
				curNode->pere->dejaTraite=true;
			}
		}
		else
			// on ajoute les fils dans l'ensemble a explorer
			for (int i=curNode->nbFils-1; i>=0; i--) nodesToExplore.push_back(curNode->fils[i]);
	}

	// initialisations des taux d'erreurs
	float genError=0.0,newGenErr=0.0;
	int nbErrTrain=launchTestTree(abr,false,false); // utile pour mise a jour incrementale
	if (!fileTest) genError=(float)(nbErrTrain+Arbre::PRUNING_FACTOR*nbLeafs)/nbLines_A;
	else genError=(float)(Arbre::WEIGHT_ERROR*nbErrTrain)/nbLines_A+(float)(launchTestTree(abr,false,true))/nbLines_T;

	// ensuite on teste chaque feuille pour voir si on ameliore l'erreur de generalisation,
	// jusqu'a ne plus rien pouvoir ameliorer.

	char memType;
	int bestIndex,max1,max2,i,memNbFils,sumRep;
	int* repartitions=new int[nbClasses]; // repartitions des classes
																// pour labeliser les noeuds
	vector<int> dejaTraites; // les id des feuilles dont des "freres" ont deja ete traites pour fusion
	while (!feuilles.empty()) {
		curNode=feuilles.back();
		feuilles.pop_back();

		// on regarde si cette feuille a des soeurs deja traitees:
		for (i=dejaTraites.size()-1; i>=0; i--)
			if (curNode->id==dejaTraites[i]) break;
		if (i>=0) continue;
		dejaTraites.push_back(curNode->id);

		// pas deja traite, donc on teste
		curNode=curNode->pere;

		// reinit de repartitions:
		for (i=nbClasses-1; i>=0; i--) repartitions[i]=0;
		sumRep=0;

		for (i=curNode->distribExemples.size()-1; i>=0; i--) {
			// au moins une init realisee
			repartitions[curNode->distribExemples[i].index]=curNode->distribExemples[i].counter;
			sumRep+=curNode->distribExemples[i].counter;
		}

		// test pour voir si le pruning est interessant
		bestIndex=nbClasses-1;
		max1=repartitions[nbClasses-1];
		max2=0;
		bool shouldPrune;
		for (i=nbClasses-2; i>=0; i--) {
			if (repartitions[i]>max1) {
				max2=max1;
				max1=repartitions[i];
				bestIndex=i;
			}
			else if (repartitions[i]>max2) max2=repartitions[i];
		}
		if ((float)max2/max1<Arbre::PRUNING_FILTER) {
			// on essaye alors la fusion:
			
			// sauvegarde de la configuration du noeud
			memType=curNode->type;
			memNbFils=curNode->nbFils;
			
			// le noeud interne devient "virtuellement" une feuille
			curNode->type='f';
			curNode->nbFils=0;
			curNode->label=initData[bestIndex].str;
			
			shouldPrune=false;
			if (!fileTest) {
				// on calcule la nouvelle erreur de generalisation
				newGenErr=(float)(nbErrTrain+sumRep-repartitions[bestIndex]+Arbre::PRUNING_FACTOR*(nbLeafs-memNbFils+1))/nbLines_A;
				if (newGenErr<=genError) shouldPrune=true;
			}
			else {
				// on calcule la nouvelle somme des taux d'erreurs
				/*** developpement futur: ponderer les termes de cette somme
				 *** afin de tenir plus compte des donnees d'entrainement ou de test,
				 *** suivant la taille de celles-ci */
				newGenErr=(float)(Arbre::WEIGHT_ERROR*(nbErrTrain+sumRep-repartitions[bestIndex]))/nbLines_A+(float)(launchTestTree(abr,false,true))/nbLines_T;
				if (newGenErr<=genError) shouldPrune=true;
			}

			if (shouldPrune) {
				// on ne restaure pas la configuration du noeud dans le cas ou la fusion est concluante
				// en revanche il faut alors ajouter le noeud aux feuilles a explorer:
				feuilles.push_back(curNode);
				nbLeafs=nbLeafs-memNbFils+1;
				genError=newGenErr;
				nbErrTrain=nbErrTrain+sumRep-repartitions[bestIndex];
				continue;
			}
			// dans le cas d'une fusion non concluante, on restaure l'etat du noeud
			curNode->type=memType;
			curNode->nbFils=memNbFils;
		}
	}
	delete [] repartitions;
}
