#include "Test.h"

// fichier de sortie contenant les taux d'erreur
ofstream reportErrors("results/errors.txt");
bool treeStats=false;

// lance le test sur le jeu de donnees souhaite, et renvoit le nombre d'erreurs
int launchTestTree(Arbre& abr,bool sideEffects,bool dataTest) {
	Noeud* curNode;
	int nbMalClasses=0,inclassables=0,index=0;
	bool missingValue; // indique si l'attribut que l'on tente de lire est manquant, auquel cas on sort de la boucle;
						// attention: ce peut etre aussi le cas d'un attribut qui a une valeur symbolique inconnue

	int nbLines=0;
	if (dataTest) nbLines=nbLines_T;
	else nbLines=nbLines_A;
	for (int i=nbLines-1; i>=0; i--) {
		curNode=abr.racine;
		missingValue=false;
		// enfin, l'exploration de l'arbre: (on s'arrte si manque la donnee de test)
		while (curNode->nbFils>0) {
			switch (curNode->type) {
				case 's':
					if (dataTest) {
						index=getIndexIn(curNode->subsets,attSymbolic_T[curNode->indexAtt][i].str);
						if (attSymbolic_T[curNode->indexAtt][i].miss || index<0) {
							missingValue=true;
							break;
						}
					}
					else {
						index=getIndexIn(curNode->subsets,attSymbolic_A[curNode->indexAtt][i].str);
						if (attSymbolic_A[curNode->indexAtt][i].miss || index<0) {
							missingValue=true;
							break;
						}
					}
					// on recherche le sous-ensemble ou se trouve la valeur d'attribut
					curNode=curNode->fils[index];
					break;
				case 'c':
					if ((!dataTest && attContinue_A[curNode->indexAtt][i].miss)
						|| (dataTest && attContinue_T[curNode->indexAtt][i].miss)) {
						missingValue=true;
						break;
					}
					if ((!dataTest && attContinue_A[curNode->indexAtt][i].val<=curNode->cutFloat)
						|| (dataTest && attContinue_T[curNode->indexAtt][i].val<=curNode->cutFloat))
						curNode=curNode->fils[0];
					else curNode=curNode->fils[1];
					break;
				case 'd':
					if ((!dataTest && attDiscrete_A[curNode->indexAtt][i].miss)
						|| (dataTest && attDiscrete_T[curNode->indexAtt][i].miss)) {
						missingValue=true;
						break;
					}
					if ((!dataTest && attDiscrete_A[curNode->indexAtt][i].val<=curNode->cutInt)
						|| (dataTest && attDiscrete_T[curNode->indexAtt][i].val<=curNode->cutInt))
						curNode=curNode->fils[0];
					else curNode=curNode->fils[1];
					break;
				default:
					curNode=curNode->fils[0]; // n'arrive jamais
					break;
			}
			if (missingValue) break;
		}
		if (curNode->type=='f') {
			// on arrive donc a une feuille (qui peut etre unknown_classe: cas particulier)
			if (curNode->label==UNKNOWN_CLASS) inclassables++;
			else if ((!dataTest && classes_A[i]!=curNode->label)
				|| (dataTest && classes_T[i]!=curNode->label)) nbMalClasses++;
		}
		else {
			// on a du arreter la recherche car valeur manquante ou non prise en compte par l'arbre,
			// on regarde si on peut raisonnablement classer le noeud
			int max1=0,max2=0,indexBest=0;
			for (int j=curNode->distribExemples.size()-1; j>=0; j--) {
				if (curNode->distribExemples[j].counter>max1) {
					max2=max1;
					max1=curNode->distribExemples[j].counter;
					indexBest=j;
				}
				else if (curNode->distribExemples[j].counter>max2)
					max2=curNode->distribExemples[j].counter;
			}
			if (max1>0 && (float)max2/max1<Arbre::PRUNING_FACTOR) {
				if ((!dataTest && classes_A[i]!=curNode->distribExemples[indexBest].str)
					|| (dataTest && classes_T[i]!=curNode->distribExemples[indexBest].str)) nbMalClasses++;
			}
			else inclassables++;
		}
	}

	if (!sideEffects) return nbMalClasses;

	// ecriture des stats dans un fichier
	if (!treeStats) {
		reportErrors << "******************************************" << endl;
		reportErrors << "*     Arbre de decision:" << endl;
		reportErrors << "******************************************" << endl << endl;
		treeStats=true;
	}
	if (dataTest) reportErrors << "  ==== Sur l'ensemble de test: ====" << endl;
	else reportErrors << "  ==== Sur l'ensemble d'entrainement: ====" << endl;
	reportErrors << "     taux d'erreur: " << (double)nbMalClasses/nbLines << endl;
	reportErrors << "     nombre d'erreurs: " << nbMalClasses;
	reportErrors << "  (dont inclassables: " << inclassables << ")" << endl;
	reportErrors << "     nombre total d'exemples: " << nbLines << endl << endl;

	return nbMalClasses;
}

void launchTestKNN(int k) {
	// parcours des donnees: on classe les donnees de test suivant leurs distances a celles d'entrainement
	string classif_knn;
	int nbMalClasses=0,inclassables=0;
	for (int i=nbLines_T-1; i>=0; i--) {
		classif_knn=classify_knn(i,k);
		if (classif_knn==UNKNOWN_CLASS) inclassables++;
		if (classif_knn!=classes_T[i]) nbMalClasses++;
	}

	// ecriture des stats dans un fichier
	reportErrors << "****************************************" << endl;
	reportErrors << "*     Classifieur " << k << "-nn:" << endl;
	reportErrors << "****************************************" << endl << endl;
	reportErrors << "     taux d'erreur: " << (double)nbMalClasses/nbLines_T << endl;
	reportErrors << "     nombre d'erreurs: " << nbMalClasses;
	reportErrors << "  (dont inclassables: " << inclassables << ")" << endl;
	reportErrors << "     nombre total d'exemples test: " << nbLines_T << endl << endl;
}

void launchTestBayes(ClassifBayes& bayes) {
	// ecriture des stats dans un fichier
	reportErrors << "**********************************************" << endl;
	reportErrors << "*     Classifieur Bayesien:" << endl;
	reportErrors << "**********************************************" << endl << endl;
	
	// parcours des donnees pour verifier le classement effectue
	string classLabel;
	int nbMalClasses=0,inclassables=0;
	for (int i=nbLines_A-1; i>=0; i--) {		classLabel=bayes.classify(i,false);
		if (classLabel==UNKNOWN_CLASS) inclassables++; // rarissime, ne devrait pas arriver
		if (classLabel!=classes_A[i]) nbMalClasses++;
	}

	reportErrors << "  ==== Sur l'ensemble d'entrainement: ====" << endl;
	reportErrors << "     taux d'erreur: " << (double)nbMalClasses/nbLines_A << endl;
	reportErrors << "     nombre d'erreurs: " << nbMalClasses;
	reportErrors << "  (dont inclassables: " << inclassables << ")" << endl;
	reportErrors << "     nombre total d'exemples: " << nbLines_A << endl << endl;

	// parcours des donnees pour verifier le classement effectue
	nbMalClasses=0;
	inclassables=0;
	for (int i=nbLines_T-1; i>=0; i--) {
		classLabel=bayes.classify(i,true);
		if (classLabel==UNKNOWN_CLASS) inclassables++; // rarissime, ne devrait pas arriver
		if (classLabel!=classes_T[i]) nbMalClasses++;
	}

	reportErrors << "  ==== Sur l'ensemble de test: ====" << endl;
	reportErrors << "     taux d'erreur: " << (double)nbMalClasses/nbLines_T << endl;
	reportErrors << "     nombre d'erreurs: " << nbMalClasses;
	reportErrors << "  (dont inclassables: " << inclassables << ")" << endl;
	reportErrors << "     nombre total d'exemples: " << nbLines_T << endl << endl;
}

void launchTestGauss(ClassifGauss& gauss) {
	// ecriture des stats dans un fichier
	reportErrors << "**********************************************" << endl;
	reportErrors << "*     Classifieur Gaussien:" << endl;
	reportErrors << "**********************************************" << endl << endl;
	
	// parcours des donnees pour verifier le classement effectue
	string classLabel;
	int nbMalClasses=0,inclassables=0;
	for (int i=nbLines_A-1; i>=0; i--) {
		classLabel=gauss.classify(i,false);
		if (classLabel==UNKNOWN_CLASS) inclassables++; // rarissime, ne devrait pas arriver
		if (classLabel!=classes_A[i]) nbMalClasses++;
	}

	reportErrors << "  ==== Sur l'ensemble d'entrainement: ====" << endl;
	reportErrors << "     taux d'erreur: " << (double)nbMalClasses/nbLines_A << endl;
	reportErrors << "     nombre d'erreurs: " << nbMalClasses;
	reportErrors << "  (dont inclassables: " << inclassables << ")" << endl;
	reportErrors << "     nombre total d'exemples: " << nbLines_A << endl << endl;

	// parcours des donnees pour verifier le classement effectue
	nbMalClasses=0;
	inclassables=0;
	for (int i=nbLines_T-1; i>=0; i--) {
		classLabel=gauss.classify(i,true);
		if (classLabel==UNKNOWN_CLASS) inclassables++; // rarissime, ne devrait pas arriver
		if (classLabel!=classes_T[i]) nbMalClasses++;
	}

	reportErrors << "  ==== Sur l'ensemble de test: ====" << endl;
	reportErrors << "     taux d'erreur: " << (double)nbMalClasses/nbLines_T << endl;
	reportErrors << "     nombre d'erreurs: " << nbMalClasses;
	reportErrors << "  (dont inclassables: " << inclassables << ")" << endl;
	reportErrors << "     nombre total d'exemples: " << nbLines_T << endl << endl;
}

void launchTestMixture(ClassifGaussMixt& gaussMixt) {
	// ecriture des stats dans un fichier
	reportErrors << "********************************************************" << endl;
	reportErrors << "*     Classifieur Gaussian Mixture:" << endl;
	reportErrors << "********************************************************" << endl << endl;
	
	// parcours des donnees pour verifier le classement effectue
	string classLabel;
	int nbMalClasses=0,inclassables=0;
	for (int i=nbLines_A-1; i>=0; i--) {
		classLabel=gaussMixt.classify(i,false);
		if (classLabel==UNKNOWN_CLASS) inclassables++; // rarissime, ne devrait pas arriver
		if (classLabel!=classes_A[i]) nbMalClasses++;
	}

	reportErrors << "  ==== Sur l'ensemble d'entrainement: ====" << endl;
	reportErrors << "     taux d'erreur: " << (double)nbMalClasses/nbLines_A << endl;
	reportErrors << "     nombre d'erreurs: " << nbMalClasses;
	reportErrors << "  (dont inclassables: " << inclassables << ")" << endl;
	reportErrors << "     nombre total d'exemples: " << nbLines_A << endl << endl;

	// parcours des donnees pour verifier le classement effectue
	nbMalClasses=0;
	inclassables=0;
	for (int i=nbLines_T-1; i>=0; i--) {
		classLabel=gaussMixt.classify(i,true);
		if (classLabel==UNKNOWN_CLASS) inclassables++; // rarissime, ne devrait pas arriver
		if (classLabel!=classes_T[i]) nbMalClasses++;
	}

	reportErrors << "  ==== Sur l'ensemble de test: ====" << endl;
	reportErrors << "     taux d'erreur: " << (double)nbMalClasses/nbLines_T << endl;
	reportErrors << "     nombre d'erreurs: " << nbMalClasses;
	reportErrors << "  (dont inclassables: " << inclassables << ")" << endl;
	reportErrors << "     nombre total d'exemples: " << nbLines_T << endl << endl;
}
