#include <string>
#include <fstream>
#include <iostream>
#include "LoadData.h"
#include "Arbre.h"
#include "ClassifTree.h"
#include "Knn.h"
#include "ClassifBayes.h"
#include "ClassifGauss.h"
#include "ClassifGaussMixt.h"
#include "Prune.h"
#include "Test.h"
#include "Timer.h"
#include <cstdlib>

using namespace std;

int main(int argc, char** argv) {

	/***************************
	* initialisations diverses
	****************************/
	
	// deux cas: tout dans le fichier argv[1], ou:
	// argv[1]: nom du fichier donnees d'entrainement
	// argv[2]: nom du fichier donnees de test
	// argv[3]: types des attributs, par exemple "idccddo"==identifiant-discret-continu-..
	// argv[4]: vecteur booleen des tests a realiser;
	//			1ere composante: arbre de decision
	//			2nde composante: test k-NN
	//			3eme composante: test baysien une gaussienne par classe
	//			4eme composante: test mixture de gaussiennes avec EM
	// argv[5]: critere de split pour la construction de l'arbre de decision
	// argv[6]: le parametre k du classifieur k-nn

	char *nameTrain,*nameTest;
	string typesAttributs;
	bool* classifieurs;
	char splitCriteria;
	int k;
	// pour mesurer le temps:
	Timer tim;
	double time0=0.0,time1=0.0,time2=0.0,time3=0.0,time4=0.0;
	
	if (argc==2) {
		ifstream dataFile(argv[1]);
		if (!dataFile.is_open()) {
			cerr << "erreur le fichier " << argv[1] << " n'existe pas" << endl;
			return 1;
		}
		string line;

		getline(dataFile,line,'\n');
		getline(dataFile,line,'\n');
		nameTrain=(char*)malloc((line.size()+1)*sizeof(char));
		for (int i=line.size()-1; i>=0; i--) nameTrain[i]=line[i];
		nameTrain[line.size()]=0;
		
		getline(dataFile,line,'\n');
		getline(dataFile,line,'\n');
		nameTest=(char*)malloc((line.size()+1)*sizeof(char));
		for (int i=line.size()-1; i>=0; i--) nameTest[i]=line[i];
		nameTest[line.size()]=0;

		getline(dataFile,line,'\n');
		getline(dataFile,typesAttributs,'\n');
		
		getline(dataFile,line,'\n');
		getline(dataFile,line,'\n');
		classifieurs=(bool*)malloc(5*sizeof(bool));
		for (int i=line.size()-1; i>=0; i--) classifieurs[i]=(line[i]!='0');
		
		char curword[10];
		getline(dataFile,line,'\n');
		dataFile >> curword;
		splitCriteria=curword[0];
		
		getline(dataFile,line,'\n');
		getline(dataFile,line,'\n');
		dataFile >> curword;
		k=atoi(curword);
	}
	else {
		nameTrain=argv[1];
		nameTest=argv[2];
		
		// on cherche le nombre et les types des attributs
		// 3 types d'attribut a gerer differemment: discret non ordonnes,continus,discret ordonnes
		// + type special: identifiant, que l'on ignore car il ne donnera aucune generalisation
		int i=0;
		while (argv[3][i]!=0) typesAttributs.push_back(argv[3][i++]);
	
		// on cherche les tests a realiser
		classifieurs=(bool*)malloc(4*sizeof(bool));
		for (i=0; i<5; i++) classifieurs[i]=(argv[4][i]-'0'!=0);

		splitCriteria=argv[5][0];
		k=atoi(argv[6]);
	}

	/******************************************
	* chargement des donnees entrainement/test
	*******************************************/

	// on charge les donnees d'entrainement
	cout << "Chargement des donnees d'entrainement" << endl;
	loadFromFile(nameTrain,typesAttributs,"train");
	// puis celles de test
	cout << "Chargement des donnees de test" << endl;
	loadFromFile(nameTest,typesAttributs,"test");

	if (classifieurs[2] || classifieurs[3] || classifieurs[4]) {
		// on recupere les ecarts-types des probabilites des donnees symboliques
		cout << "Recherche des probabilites des donnees symboliques" << endl;
		getDistribSymbolic();
	}
	if (classifieurs[1] || classifieurs[2] || classifieurs[3] || classifieurs[4]) {
		// normalisation des donnees numeriques:
		cout << "Normalisation des donnees numeriques" << endl;
		normalizeData();
	}

	/*************************************
	* 1) creation de l'arbre de decision
	**************************************/

	Arbre abr;
	if (classifieurs[0]) {
		cout << "Creation de l'arbre de decision" << endl;
		tim.restart();
		// on passe alors en parametre a la fonction de creation de l'arbre le type de split desire
		abr=buildTree(splitCriteria);
		// on coupe les branches de l'arbre trop specialisees pour gagner en generalisation
		pruning(abr,false);
		pruning(abr,true);
		time0+=tim.elapsed();
	}

	/********************************************
	* 2) creation du classificateur k-NN
	* [ rien a faire, c'est juste une fonction ]
	*********************************************/
	
	/********************************************
	* 3) creation du classifieur bayesien simple
	*********************************************/

	ClassifBayes bayes;
	if (classifieurs[2]) {
		cout << "Creation du classifieur Bayes" << endl;
		tim.restart();
		bayes=ClassifBayes(typesAttributs);
		bayes.findParams();
		time2+=tim.elapsed();
	}

	/****************************************************
	* 4) creation du classifieur gaussian multi-dim
	*    [ne presuppose pas l'independance des attributs]
	*****************************************************/

	ClassifGauss gauss;
	if (classifieurs[3]) {
		cout << "Creation du classifieur Gauss" << endl;
		tim.restart();
		gauss=ClassifGauss(typesAttributs);
		gauss.initialize();
		time3+=tim.elapsed();
	}

	/*********************************************
	* 5) creation du classifieur gaussian mixture
	**********************************************/

	ClassifGaussMixt gaussMixt;
	if (classifieurs[4]) {
		cout << "Creation du classifieur GaussMixt" << endl;
		tim.restart();
		gaussMixt=ClassifGaussMixt(typesAttributs);
		gaussMixt.initialize();
		time4+=tim.elapsed();
	}
	
	/*******************************************
	* TESTS de tous les classifieurs construits
	********************************************/

	if (classifieurs[0]) {
		cout << "Test de l'arbre de decision" << endl;
		tim.restart();
		launchTestTree(abr,true,false);
		launchTestTree(abr,true,true); // true pour dire qu'on veut des effets de bord dans un fichier,
										// et sur quel ensemble on teste
		time0+=tim.elapsed();
	}
	if (classifieurs[1]) {		
		cout << "Test du classifieur K-nn" << endl;
		tim.restart();
		launchTestKNN(k);
		time1+=tim.elapsed();
	}
	if (classifieurs[2]) {
		cout << "Test du classifieur Bayes" << endl;
		tim.restart();
		launchTestBayes(bayes);
		time2+=tim.elapsed();
	}
	if (classifieurs[3]) {
		cout << "Test du classifieur Gauss" << endl;
		tim.restart();
		launchTestGauss(gauss);
		time3+=tim.elapsed();
	}
	if (classifieurs[4]) {
		cout << "Test du classifieur GaussMixt" << endl;
		tim.restart();
		launchTestMixture(gaussMixt);
		time4+=tim.elapsed();
	}

	/*********************************************************************
	* visualisation des classifieurs construits;
	* pour Bayes, gaussian et EM on stocke les parametres dans un fichier
	**********************************************************************/
	
	if (classifieurs[0]) {
		// visualisation de l'arbre calcule
		abr.visualize();
	}
	if (classifieurs[1]) {
		// puis du classifieur k-NN
		// eventuellement on pourrait tracer la frontiere sur un sous-espace de dimension 2;
		// a voir plus tard.
	}
	if (classifieurs[2]) {
		// puis du classifieur bayesien
		bayes.printToFile();
	}
	if (classifieurs[3]) {
		// puis du classifieur bayesien
		gauss.printToFile();
	}
	if (classifieurs[4]) {
		// puis du classifieur gaussian mixture
		gaussMixt.printToFile();
	}
	
	cout << endl << "Temps totaux des divers classifieurs (creation + tests):" << endl;
	if (classifieurs[0]) cout << "Arbre de decision: " << time0 << endl;
	if (classifieurs[1]) cout << "Classifieur K-nn: " << time1 << endl;
	if (classifieurs[2]) cout << "Classifieur Bayes: " << time2 << endl;
	if (classifieurs[3]) cout << "Classifieur Gauss: " << time3 << endl;
	if (classifieurs[4]) cout << "Classifieur GaussMixt: " << time4 << endl;

	return 0;
}
